
#include "rt2870_softc.h"
#include "rt2870_reg.h"
#include "rt2870_eeprom.h"
#include "rt2870_ucode.h"
#include "rt2870_rxwi.h"
#include "rt2870_rxinfo.h"
#include "rt2870_txwi.h"
#include "rt2870_txinfo.h"
#include "rt2870_read_eeprom.h"
#include "rt2870_io.h"
#include "rt2870_rf.h"
#include "rt2870_led.h"
#include "rt2870_debug.h"

/*
 * Defines and macros
 */

#define USB_PRODUCT_LINKSYS4_WUSB600N			0x0071
#define USB_PRODUCT_DLINK2_DWA140				0x3c09
#define USB_PRODUCT_DLINK2_DWA160AREVB			0x3c11

#define RT2870_USB_CONFIG_NO					1
#define RT2870_USB_IFACE_INDEX					0

#define RT2870_USB_REQ_MAC_READ_MULTI			0x07

/* packet length + Rx wireless info + Rx info */
#define RT2870_RX_DESC_SIZE						\
	(sizeof(uint32_t) + sizeof(struct rt2870_rxwi) + sizeof(struct rt2870_rxinfo))

/* Tx info + Tx wireless info + max padding */
#define RT2870_TX_DESC_SIZE						\
	(sizeof(struct rt2870_txinfo) + sizeof(struct rt2870_txwi) + 11)

#define RT2870_MAX_AGG_SIZE						3840

#define RT2870_USB_RX_BULK_BUFLEN				(2048 * 12)

#define RT2870_NOISE_FLOOR						-95

#define RT2870_AID2WCID(aid)					((aid) & 0xff)

#define RT2870_RATE_IS_OFDM(rate)				((rate) >= 12 && (rate) != 22)

#define RT2870_ACK_SIZE							14

#define RT2870_TX_TIMEOUT						5000

#define IEEE80211_HAS_ADDR4(wh)					\
	(((wh)->i_fc[1] & IEEE80211_FC1_DIR_MASK) == IEEE80211_FC1_DIR_DSTODS)

/*
 * Data structures and types
 */

struct rt2870_cmd_argv_newstate
{
	enum ieee80211_state state;
	int arg;
};

struct rt2870_cmd_argv_newassoc
{
	uint16_t associd;
	int isnew;
	uint8_t macaddr[IEEE80211_ADDR_LEN];
};

struct rt2870_cmd_argv_updatebeacon
{
	int what;
};

struct rt2870_cmd_argv_keyset
{
	struct ieee80211_key key;
	uint16_t associd;
};

struct rt2870_cmd_argv_keydelete
{
	struct ieee80211_key key;
	uint16_t associd;
};

/*
 * Static function prototypes
 */

static int rt2870_find_usb_endpoints(struct rt2870_softc *sc);

static void rt2870_init(void *priv);

static int rt2870_init_bbp(struct rt2870_softc *sc);

static void rt2870_stop(void *priv);

static void rt2870_start(struct ifnet *ifp);

static int rt2870_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data);

static int rt2870_newstate(struct ieee80211com *ic,
	enum ieee80211_state nstate, int arg);

static void rt2870_scan_start(struct ieee80211com *ic);

static void rt2870_scan_end(struct ieee80211com *ic);

static void rt2870_set_channel(struct ieee80211com *ic);

static void rt2870_newassoc(struct ieee80211_node *ni, int isnew);

static void rt2870_updateslot(struct ifnet *ifp);

static int rt2870_wme_update(struct ieee80211com *ic);

static void rt2870_update_beacon(struct ieee80211com *ic, int what);

static int rt2870_key_set(struct ieee80211com *ic,
	const struct ieee80211_key *k, const uint8_t mac[IEEE80211_ADDR_LEN]);

static int rt2870_key_delete(struct ieee80211com *ic,
	const struct ieee80211_key *k);

static int rt2870_raw_xmit(struct ieee80211_node *ni, struct mbuf *m,
	const struct ieee80211_bpf_params *params);

static int rt2870_media_change(struct ifnet *ifp);

static void rt2870_amrr_start(struct rt2870_softc *sc,
	struct ieee80211_node *ni);

static void rt2870_amrr_timeout(void *arg);

static void rt2870_amrr_update(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status);

static void rt2870_tx_watchdog(void *arg);

static void rt2870_adjust_txpow(void *arg);

static void rt2870_newstate_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_scan_start_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_scan_end_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_set_channel_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_newassoc_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_updateslot_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_update_beacon_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_wme_update_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_key_set_cb(struct rt2870_softc *sc, void *arg);

static void rt2870_key_delete_cb(struct rt2870_softc *sc, void *arg);

static int rt2870_do_async(struct rt2870_softc *sc,
	void (*cb)(struct rt2870_softc *sc, void *priv),
	void *arg, int len);

static void rt2870_cmd_task(void *arg);

static void rt2870_asic_set_bssid(struct rt2870_softc *sc,
	const uint8_t *bssid);

static void rt2870_asic_set_macaddr(struct rt2870_softc *sc,
	const uint8_t *addr);

static void rt2870_asic_enable_tsf_sync(struct rt2870_softc *sc);

static void rt2870_asic_disable_tsf_sync(struct rt2870_softc *sc);

static void rt2870_asic_update_promisc(struct rt2870_softc *sc);

static void rt2870_asic_enable_mrr(struct rt2870_softc *sc);

static void rt2870_asic_set_txpreamble(struct rt2870_softc *sc);

static void rt2870_asic_set_basicrates(struct rt2870_softc *sc);

static void rt2870_asic_update_rtsthreshold(struct rt2870_softc *sc);

static void rt2870_asic_updateprot(struct rt2870_softc *sc);

static void rt2870_asic_updateslot(struct rt2870_softc *sc);

static void rt2870_asic_wme_update(struct rt2870_softc *sc);

static void rt2870_asic_update_beacon(struct rt2870_softc *sc);

static void rt2870_asic_clear_keytables(struct rt2870_softc *sc);

static uint8_t rt2870_rxrate(struct rt2870_rxwi *rxwi);

static uint8_t rt2870_maxrssi_rxpath(struct rt2870_softc *sc,
	const struct rt2870_rxwi *rxwi);

static int8_t rt2870_rssi2dbm(struct rt2870_softc *sc,
	uint8_t rssi, uint8_t rxpath);

static uint8_t rt2870_rate2mcs(uint8_t rate);

static int rt2870_ackrate(struct ieee80211com *ic, int rate);

static uint16_t rt2870_txtime(int len, int rate, uint32_t flags);

static void rt2870_rx_intr(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status);

static void rt2870_rx_frame(struct rt2870_softc *sc,
	uint8_t *buf, uint32_t dmalen);

static int rt2870_tx_frame(struct rt2870_softc *sc,
	struct mbuf *m, struct ieee80211_node *ni, int qid);

static int rt2870_tx_raw(struct rt2870_softc *sc,
	struct mbuf *m, struct ieee80211_node *ni,
	const struct ieee80211_bpf_params *params);

static void rt2870_tx_intr(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status);

static int rt2870_alloc_rx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_rx_ring *ring);

static void rt2870_free_rx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_rx_ring *ring);

static int rt2870_alloc_tx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_tx_ring *ring, int qid);

static void rt2870_free_tx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_tx_ring *ring);

/*
 * Static variables
 */

static const struct usb_devno rt2870_usb_devno[] =
{
	{ USB_VENDOR_LINKSYS4, USB_PRODUCT_LINKSYS4_WUSB600N },		/* Linksys WUSB600N */
	{ USB_VENDOR_DLINK2, USB_PRODUCT_DLINK2_DWA140 },			/* D-Link DWA-140 */
	{ USB_VENDOR_DLINK2, USB_PRODUCT_DLINK2_DWA160AREVB },		/* D-Link DWA-160A Rev. B */
};

static const struct
{
	uint32_t reg;
	uint32_t val;
} rt2870_def_mac[] =
{
	{ RT2870_REG_PBF_BCN_OFFSET0, 			0xf8f0e8e0 },
	{ RT2870_REG_PBF_BCN_OFFSET1, 			0x6f77d0c8 },
	{ RT2870_REG_LEGACY_BASIC_RATE, 		0x0000013f },
	{ RT2870_REG_HT_BASIC_RATE, 			0x00008003 },
	{ RT2870_REG_SYS_CTRL, 					0x00000000 },
	{ RT2870_REG_RX_FILTER_CFG, 			0x00017f97 },
	{ RT2870_REG_BKOFF_SLOT_CFG, 			0x00000209 },
	{ RT2870_REG_TX_SW_CFG0, 				0x00000000 },
	{ RT2870_REG_TX_SW_CFG1, 				0x00080606 },
	{ RT2870_REG_TX_LINK_CFG, 				0x00001020 },
	{ RT2870_REG_TX_TIMEOUT_CFG, 			0x000a2090 },
	{ RT2870_REG_MAX_LEN_CFG, 				(1 << 12) | RT2870_MAX_AGG_SIZE },
	{ RT2870_REG_LED_CFG, 					0x7f031e46 },
	{ RT2870_REG_PBF_MAX_PCNT, 				0x1f3fbf9f },
	{ RT2870_REG_TX_RTY_CFG, 				0x47d01f0f },
	{ RT2870_REG_AUTO_RSP_CFG, 				0x00000013 },
	{ RT2870_REG_TX_CCK_PROT_CFG, 			0x05740003 },
	{ RT2870_REG_TX_OFDM_PROT_CFG, 			0x05740003 },
	{ RT2870_REG_PBF_CFG, 					0x00f40006 },
	{ RT2870_REG_TX_MM40_PROT_CFG, 			0x03f44084 },
	{ RT2870_REG_SCHDMA_WPDMA_GLO_CFG, 		0x00000030 },
	{ RT2870_REG_TX_GF20_PROT_CFG, 			0x01744004 },
	{ RT2870_REG_TX_GF40_PROT_CFG, 			0x03f44084 },
	{ RT2870_REG_TX_MM20_PROT_CFG, 			0x01744004 },
	{ RT2870_REG_TX_TXOP_CTRL_CFG, 			0x0000583f },
	{ RT2870_REG_TX_RTS_CFG, 				0x00092b20 },
	{ RT2870_REG_TX_EXP_ACK_TIME, 			0x002400ca },
	{ RT2870_REG_HCCAPSMP_TXOP_HLDR_ET, 	0x00000002 },
	{ RT2870_REG_XIFS_TIME_CFG, 			0x33a41010 },
	{ RT2870_REG_PWR_PIN_CFG, 				0x00000003 },
	{ RT2870_REG_SCHDMA_WMM_AIFSN_CFG,		0x00002273 },
	{ RT2870_REG_SCHDMA_WMM_CWMIN_CFG,		0x00002344 },
	{ RT2870_REG_SCHDMA_WMM_CWMAX_CFG,		0x000034aa },
};

#define RT2870_DEF_MAC_SIZE		(sizeof(rt2870_def_mac) / sizeof(rt2870_def_mac[0]))

static const struct
{
	uint8_t	reg;
	uint8_t	val;
} rt2870_def_bbp[] =
{
	{ 65,	0x2c },
	{ 66,	0x38 },
	{ 69,	0x12 },
	{ 70,	0x0a },
	{ 73,	0x10 },
	{ 81,	0x37 },
	{ 82,	0x62 },
	{ 83,	0x6a },
	{ 84,	0x99 },
	{ 86,	0x00 },
	{ 91,	0x04 },
	{ 92,	0x00 },
	{ 103,	0x00 },
	{ 105,	0x05 },
};

#define RT2870_DEF_BBP_SIZE		(sizeof(rt2870_def_bbp) / sizeof(rt2870_def_bbp[0]))

/*
 * rt2870_probe
 */
static int rt2870_probe(device_t dev)
{
	struct usb_attach_arg *uaa;

	uaa = device_get_ivars(dev);

	if (uaa->iface != NULL)
		return UMATCH_NONE;

	return (usb_lookup(rt2870_usb_devno, uaa->vendor, uaa->product) != NULL) ?
		UMATCH_VENDOR_PRODUCT : UMATCH_NONE;
}

/*
 * rt2870_attach
 */
static int rt2870_attach(device_t dev)
{
	struct rt2870_softc *sc;
	struct usb_attach_arg *uaa;
	usbd_status usb_error;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	int error, ntries, bands, i;

	sc = device_get_softc(dev);
	uaa = device_get_ivars(dev);

	sc->dev = dev;
	sc->usb_dev = uaa->device;

#ifdef RT2870_DEBUG
	sc->debug = 1;

	SYSCTL_ADD_INT(device_get_sysctl_ctx(dev),
		SYSCTL_CHILDREN(device_get_sysctl_tree(dev)), OID_AUTO,
		"debug", CTLFLAG_RW, &sc->debug, 0, "rt2870 debug level");
#endif

	if (usbd_set_config_no(sc->usb_dev, RT2870_USB_CONFIG_NO, 0) != 0)
	{
		printf("%s: could not set usb configuration no\n",
			device_get_nameunit(sc->dev));
		return ENXIO;
	}

	usb_error = usbd_device2interface_handle(sc->usb_dev,
		RT2870_USB_IFACE_INDEX, &sc->usb_iface);
	if (usb_error != 0)
	{
		printf("%s: could not get usb interface handle\n",
			device_get_nameunit(sc->dev));
		return ENXIO;
	}

	error = rt2870_find_usb_endpoints(sc);
	if (error != 0)
		return error;

	for (ntries = 0; ntries < 100; ntries++)
	{
		sc->mac_rev = rt2870_io_mac_read(sc, RT2870_REG_MAC_CSR0);
		if (sc->mac_rev != 0x00000000 && sc->mac_rev != 0xffffffff)
			break;

		DELAY(10);
	}

	if (ntries == 100)
	{
		printf("%s: timeout waiting for NIC to initialize\n",
			device_get_nameunit(sc->dev));
		return ENXIO;
	}

	rt2870_read_eeprom(sc);

	printf("%s: MAC/BBP RT2870 (rev 0x%08x), RF %s\n",
	    device_get_nameunit(sc->dev), sc->mac_rev,
		rt2870_rf_name(sc->rf_rev));

	/* allocate xfer for AMRR statistics requests */

	sc->amrr_xfer = usbd_alloc_xfer(sc->usb_dev);
	if (sc->amrr_xfer == NULL)
	{
		printf("%s: could not allocate AMRR xfer\n",
		    device_get_nameunit(sc->dev));
		error = ENOMEM;
		goto fail;
	}

	/* open Tx and Rx USB bulk pipes */

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
	{
		error = usbd_open_pipe(sc->usb_iface, sc->tx_ring[i].usb_ep,
			USBD_EXCLUSIVE_USE, &sc->tx_ring[i].usb_pipe);
		if (error != 0)
		{
			printf("%s: could not open Tx pipe #%d: %s\n",
				device_get_nameunit(sc->dev), i, usbd_errstr(error));
			goto fail;
		}
	}

	error = usbd_open_pipe(sc->usb_iface, sc->rx_ring.usb_ep,
		USBD_EXCLUSIVE_USE, &sc->rx_ring.usb_pipe);
	if (error != 0)
	{
		printf("%s: could not open Rx pipe: %s\n",
			device_get_nameunit(sc->dev), usbd_errstr(error));
		goto fail;
	}

	/* allocate Tx and Rx rings */

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
	{
		error = rt2870_alloc_tx_ring(sc, &sc->tx_ring[i], i);
		if (error != 0)
		{
			printf("%s: could not allocate Tx ring #%d\n",
				device_get_nameunit(sc->dev), i);
			goto fail;
		}
	}

	sc->tx_ring_mgtqid = WME_AC_BE;

	error = rt2870_alloc_rx_ring(sc, &sc->rx_ring);
	if (error != 0)
	{
		printf("%s: could not allocate Rx ring\n",
			device_get_nameunit(sc->dev));
		goto fail;
	}

	callout_init(&sc->amrr_ch, 0);
	callout_init(&sc->tx_watchdog_ch, 0);
	callout_init(&sc->adjust_txpow_ch, 0);

	usb_init_task(&sc->cmd_task, rt2870_cmd_task, sc);

	ifp = sc->ifp = if_alloc(IFT_ETHER);
	if (ifp == NULL)
	{
		printf("%s: could not if_alloc()\n",
			device_get_nameunit(sc->dev));
		error = ENOMEM;
		goto fail;
	}

	ifp->if_softc = sc;

	if_initname(ifp, "rt2870", device_get_unit(sc->dev));

	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST |
		IFF_NEEDSGIANT;

	ifp->if_init = rt2870_init;
	ifp->if_ioctl = rt2870_ioctl;
	ifp->if_start = rt2870_start;

	IFQ_SET_MAXLEN(&ifp->if_snd, IFQ_MAXLEN);
	ifp->if_snd.ifq_drv_maxlen = IFQ_MAXLEN;
	IFQ_SET_READY(&ifp->if_snd);

	ic = &sc->ic;

	ic->ic_ifp = ifp;

	ic->ic_phytype = IEEE80211_T_OFDM;
	ic->ic_opmode = IEEE80211_M_STA;
	ic->ic_state = IEEE80211_S_INIT;

	ic->ic_caps = IEEE80211_C_MONITOR |
		IEEE80211_C_IBSS |
		IEEE80211_C_HOSTAP |
		IEEE80211_C_BGSCAN |
		IEEE80211_C_TXPMGT |
	    IEEE80211_C_SHPREAMBLE |
	    IEEE80211_C_SHSLOT |
		IEEE80211_C_WME |
		IEEE80211_C_WEP |
		IEEE80211_C_TKIP |
		IEEE80211_C_TKIPMIC |
		IEEE80211_C_AES_CCM |
		IEEE80211_C_WPA;

	bands = 0;

	setbit(&bands, IEEE80211_MODE_11B);
	setbit(&bands, IEEE80211_MODE_11G);

	if (sc->rf_rev == RT2870_EEPROM_RF_2850 ||
		sc->rf_rev == RT2870_EEPROM_RF_2750)
		setbit(&bands, IEEE80211_MODE_11A);

	ieee80211_init_channels(ic, 0, CTRY_DEFAULT, bands, 0, 1);

	ieee80211_ifattach(ic);

	sc->newstate = ic->ic_newstate;
	ic->ic_newstate = rt2870_newstate;

	ic->ic_scan_start = rt2870_scan_start;
	ic->ic_scan_end = rt2870_scan_end;
	ic->ic_set_channel = rt2870_set_channel;
	ic->ic_newassoc = rt2870_newassoc;
	ic->ic_updateslot = rt2870_updateslot;
	ic->ic_wme.wme_update = rt2870_wme_update;
	ic->ic_update_beacon = rt2870_update_beacon;
	ic->ic_crypto.cs_key_set = rt2870_key_set;
	ic->ic_crypto.cs_key_delete = rt2870_key_delete;
	ic->ic_raw_xmit = rt2870_raw_xmit;

	ic->ic_flags_ext |= IEEE80211_FEXT_SWBMISS;

	ieee80211_media_init(ic, rt2870_media_change, ieee80211_media_status);

	ieee80211_amrr_init(&sc->amrr, ic,
		IEEE80211_AMRR_MIN_SUCCESS_THRESHOLD,
		IEEE80211_AMRR_MAX_SUCCESS_THRESHOLD);

	bpfattach2(ifp, DLT_IEEE802_11_RADIO,
	    sizeof(struct ieee80211_frame) + IEEE80211_RADIOTAP_HDRLEN, 
	    &sc->drvbpf);

	sc->rxtap_len = sizeof(sc->rxtapu);
	sc->rxtap.ihdr.it_len = htole16(sc->rxtap_len);
	sc->rxtap.ihdr.it_present = htole32(RT2870_SOFTC_RX_RADIOTAP_PRESENT);

	sc->txtap_len = sizeof(sc->txtapu);
	sc->txtap.ihdr.it_len = htole16(sc->txtap_len);
	sc->txtap.ihdr.it_present = htole32(RT2870_SOFTC_TX_RADIOTAP_PRESENT);

	if (bootverbose)
		ieee80211_announce(ic);

	usbd_add_drv_event(USB_EVENT_DRIVER_ATTACH, sc->usb_dev, sc->dev);

	return 0;

fail:

	if (sc->amrr_xfer != NULL)
	{
		usbd_free_xfer(sc->amrr_xfer);
		sc->amrr_xfer = NULL;
	}

	/* close Tx and Rx USB bulk pipes */

	if (sc->rx_ring.usb_pipe != NULL)
	{
		usbd_abort_pipe(sc->rx_ring.usb_pipe);
		usbd_close_pipe(sc->rx_ring.usb_pipe);
		sc->rx_ring.usb_pipe = NULL;
	}

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
	{
		if (sc->tx_ring[i].usb_pipe != NULL)
		{
			usbd_abort_pipe(sc->tx_ring[i].usb_pipe);
			usbd_close_pipe(sc->tx_ring[i].usb_pipe);
			sc->tx_ring[i].usb_pipe = NULL;
		}
	}

	/* free Tx and Rx rings */

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
		rt2870_free_tx_ring(sc, &sc->tx_ring[i]);
	
	rt2870_free_rx_ring(sc, &sc->rx_ring);

	return error;
}

/*
 * rt2870_detach
 */
static int rt2870_detach(device_t dev)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	int i;

	if (!device_is_attached(dev))
		return 0;

	sc = device_get_softc(dev);
	ic = &sc->ic;
	ifp = ic->ic_ifp;

	sc->tx_timer = 0;

	ifp->if_drv_flags &= ~(IFF_DRV_RUNNING | IFF_DRV_OACTIVE);

	ieee80211_new_state(ic, IEEE80211_S_INIT, -1);

	usb_rem_task(sc->usb_dev, &sc->cmd_task);

	callout_stop(&sc->amrr_ch);
	callout_stop(&sc->tx_watchdog_ch);
	callout_stop(&sc->adjust_txpow_ch);

	if (sc->amrr_xfer != NULL)
	{
		usbd_free_xfer(sc->amrr_xfer);
		sc->amrr_xfer = NULL;
	}

	/* close Tx and Rx USB bulk pipes */

	if (sc->rx_ring.usb_pipe != NULL)
	{
		usbd_abort_pipe(sc->rx_ring.usb_pipe);
		usbd_close_pipe(sc->rx_ring.usb_pipe);
		sc->rx_ring.usb_pipe = NULL;
	}

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
	{
		if (sc->tx_ring[i].usb_pipe != NULL)
		{
			usbd_abort_pipe(sc->tx_ring[i].usb_pipe);
			usbd_close_pipe(sc->tx_ring[i].usb_pipe);
			sc->tx_ring[i].usb_pipe = NULL;
		}
	}

	/* free Tx and Rx rings */

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
		rt2870_free_tx_ring(sc, &sc->tx_ring[i]);
	
	rt2870_free_rx_ring(sc, &sc->rx_ring);

	bpfdetach(ifp);

	ieee80211_ifdetach(ic);

	if_free(ifp);

	usbd_add_drv_event(USB_EVENT_DRIVER_DETACH, sc->usb_dev, sc->dev);

	return 0;
}

/*
 * rt2870_find_usb_endpoints
 */
static int rt2870_find_usb_endpoints(struct rt2870_softc *sc)
{
	usb_interface_descriptor_t *id;
	usb_endpoint_descriptor_t *ed;
	int i, j;

	id = usbd_get_interface_descriptor(sc->usb_iface);

	sc->rx_ring.usb_ep = -1;

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
		sc->tx_ring[i].usb_ep = -1;

	for (i = 0, j = 0; i < id->bNumEndpoints; i++)
	{
		ed = usbd_interface2endpoint_descriptor(sc->usb_iface, i);
		if (ed == NULL)
		{
			printf("%s: no endpoint descriptor #%d for usb interface\n",
				device_get_nameunit(sc->dev), i);
			return ENXIO;
		}

		if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_IN &&
			UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK)
		{
			sc->rx_ring.usb_ep = ed->bEndpointAddress;
		}
		else if (UE_GET_DIR(ed->bEndpointAddress) == UE_DIR_OUT &&
			UE_GET_XFERTYPE(ed->bmAttributes) == UE_BULK)
		{
			if (j < RT2870_SOFTC_TX_RING_COUNT)
				sc->tx_ring[j++].usb_ep = ed->bEndpointAddress;
			else
				sc->tx_ring[RT2870_SOFTC_TX_RING_COUNT - 1].usb_ep =
					ed->bEndpointAddress;
		}
	}

	if (sc->rx_ring.usb_ep == -1 || j < RT2870_SOFTC_TX_RING_COUNT)
	{
		printf("%s: missing usb endpoints\n",
			device_get_nameunit(sc->dev));
		return ENXIO;
	}

	return 0;
}

/*
 * rt2870_init
 */
static void rt2870_init(void *priv)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct rt2870_softc_rx_data *data;
	int ntries, error, i;
	uint32_t tmp;

	sc = priv;
	ic = &sc->ic;
	ifp = ic->ic_ifp;

	if (!(sc->flags & RT2870_SOFTC_FLAGS_UCODE_LOADED))
	{
		DPRINTF(sc, ("%s: loading 8051 microcode\n",
			device_get_nameunit(sc->dev)));

		error = rt2870_io_mcu_load_ucode(sc, rt2870_ucode, sizeof(rt2870_ucode));
		if (error != 0)
		{
			printf("%s: could not load 8051 microcode\n",
				device_get_nameunit(sc->dev));
			goto fail;
		}

		DPRINTF(sc, ("%s: 8051 microcode was successfully loaded\n",
			device_get_nameunit(sc->dev)));

		sc->flags |= RT2870_SOFTC_FLAGS_UCODE_LOADED;
	}

	/* wait while DMA engine is busy */

	for (ntries = 0; ntries < 100; ntries++)
	{
		tmp = rt2870_io_mac_read(sc, RT2870_REG_SCHDMA_WPDMA_GLO_CFG);
		if (!(tmp & (RT2870_REG_TX_DMA_BUSY | RT2870_REG_RX_DMA_BUSY)))
			break;

		DELAY(1000);
	}

	if (ntries == 100)
	{
		printf("%s: timeout waiting for DMA engine\n",
			device_get_nameunit(sc->dev));
		goto fail;
	}

	tmp &= 0xff0;
	tmp |= RT2870_REG_TX_WB_DDONE;

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WPDMA_GLO_CFG, tmp);

	/* PBF hardware reset */

	tmp = rt2870_io_mac_read(sc, RT2870_REG_PBF_SYS_CTRL);

	tmp &= ~(1 << 13);

	rt2870_io_mac_write(sc, RT2870_REG_PBF_SYS_CTRL, tmp);

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL,
		RT2870_REG_MAC_SRST | RT2870_REG_BBP_HRST);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_USB_DMA_CFG, 0);

	rt2870_io_mcu_reset(sc);

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL, 0);

	/* init Tx power compensation */

	for (i = 0; i < RT2870_SOFTC_TXPOW_COMP_COUNT; i++)
	{
		if (sc->txpow_comp_20mhz[i] == 0xffffffff)
			continue;

		rt2870_io_mac_write(sc, RT2870_REG_TX_PWR_CFG(i),
			sc->txpow_comp_20mhz[i]);
	}

	for (i = 0; i < RT2870_DEF_MAC_SIZE; i++)
		rt2870_io_mac_write(sc, rt2870_def_mac[i].reg,
			rt2870_def_mac[i].val);

	/* wait while MAC is busy */

	for (ntries = 0; ntries < 100; ntries++)
	{
		if (!(rt2870_io_mac_read(sc, RT2870_REG_STATUS_CFG) &
			(RT2870_REG_STATUS_TX_BUSY | RT2870_REG_STATUS_RX_BUSY)))
			break;

		DELAY(1000);
	}

	if (ntries == 100)
	{
		printf("%s: timeout waiting for MAC\n",
			device_get_nameunit(sc->dev));
		goto fail;
	}

	/* clear Host to MCU mailbox */

	rt2870_io_mac_write(sc, RT2870_REG_H2M_MAILBOX_BBP_AGENT, 0);
	rt2870_io_mac_write(sc, RT2870_REG_H2M_MAILBOX, 0);

	DELAY(1000);

	error = rt2870_init_bbp(sc);
	if (error != 0)
		goto fail;

	/* set up maximum buffer sizes */

	tmp = (1 << 12) | RT2870_MAX_AGG_SIZE;

	if (sc->mac_rev >= 0x28720200 && sc->mac_rev < 0x30700200)
	{
		tmp &= 0xfff;
		tmp |= 0x2000;
	}

	rt2870_io_mac_write(sc, RT2870_REG_MAX_LEN_CFG, tmp);

	/* clear STA registers */

	rt2870_io_mac_read_multi(sc, RT2870_REG_RX_STA_CNT0,
		sc->amrr_sta, sizeof(sc->amrr_sta));

	/* set mac address */

	IEEE80211_ADDR_COPY(ic->ic_myaddr, IF_LLADDR(ifp));

	rt2870_asic_set_macaddr(sc, ic->ic_myaddr);

	/* send LEDs operating mode to microcontroller */

	rt2870_io_mcu_cmd(sc, RT2870_IO_MCU_CMD_LED1, sc->led_off[0]);
	rt2870_io_mcu_cmd(sc, RT2870_IO_MCU_CMD_LED2, sc->led_off[1]);
	rt2870_io_mcu_cmd(sc, RT2870_IO_MCU_CMD_LED3, sc->led_off[2]);

	/* write vendor-specific BBP values (from EEPROM) */

	for (i = 0; i < RT2870_SOFTC_BBP_EEPROM_COUNT; i++)
	{
		if (sc->bbp_eeprom[i].reg == 0x00 ||
			sc->bbp_eeprom[i].reg == 0xff)
			continue;

		rt2870_io_bbp_write(sc, sc->bbp_eeprom[i].reg,
			sc->bbp_eeprom[i].val);
	}

	/* disable non-existing Rx chains */

	tmp = rt2870_io_bbp_read(sc, 3);

	tmp &= ~((1 << 4) | (1 << 3));

	if (sc->nrxpath == 3)
		tmp |= (1 << 4);
	else if (sc->nrxpath == 2)
		tmp |= (1 << 3);

	rt2870_io_bbp_write(sc, 3, tmp);

	/* disable non-existing Tx chains */

	tmp = rt2870_io_bbp_read(sc, 1);

	if (sc->ntxpath == 1)
		tmp &= ~((1 << 4) | (1 << 3));

	rt2870_io_bbp_write(sc, 1, tmp);

	tmp = rt2870_io_bbp_read(sc, 4);

	tmp &= ~0x18;

	rt2870_io_bbp_write(sc, 4, tmp);

	/* set current channel */

	rt2870_rf_set_chan(sc, ic->ic_curchan);

	/* turn radio LED on */

	rt2870_led_cmd(sc, RT2870_LED_CMD_RADIO_ON);

	rt2870_io_mcu_cmd(sc, RT2870_IO_MCU_CMD_BOOT, 0);

	/* set RTS threshold */

	rt2870_asic_update_rtsthreshold(sc);

	/* set up protection mode */

	rt2870_asic_updateprot(sc);

	/* clear key tables */

	rt2870_asic_clear_keytables(sc);

	/* enable Tx/Rx DMA engine */

	tmp = rt2870_io_mac_read(sc, RT2870_REG_SCHDMA_USB_CYC_CFG);

	tmp &= 0xffffff00;
	tmp |= 0x1e;

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_USB_CYC_CFG, tmp);

	if ((sc->mac_rev & 0xffff) != 0x0101)
		rt2870_io_mac_write(sc, RT2870_REG_TX_TXOP_CTRL_CFG, 0x583f);

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL, RT2870_REG_TX_ENABLE);

	for (ntries = 0; ntries < 200; ntries++)
	{
		tmp = rt2870_io_mac_read(sc, RT2870_REG_SCHDMA_WPDMA_GLO_CFG);
		if (!(tmp & (RT2870_REG_TX_DMA_BUSY | RT2870_REG_RX_DMA_BUSY)))
			break;

		DELAY(1000);
	}

	if (ntries == 200)
	{
		printf("%s: timeout waiting for DMA engine\n",
			device_get_nameunit(sc->dev));
		goto fail;
	}

	DELAY(50);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_TXOP0_CFG, 0);
	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_TXOP1_CFG,
		(48 << 16) | 96);

	tmp |= RT2870_REG_TX_WB_DDONE |
		RT2870_REG_RX_DMA_ENABLE |
		RT2870_REG_TX_DMA_ENABLE;

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WPDMA_GLO_CFG, tmp);

	tmp = RT2870_REG_USB_DMA_TX_ENABLE |
		RT2870_REG_USB_DMA_RX_ENABLE |
		RT2870_REG_USB_DMA_RX_AGG_ENABLE |
		/* Rx agg limit in unit of 1024 byte */
		((RT2870_USB_RX_BULK_BUFLEN / 1024 - 3) << RT2870_REG_USB_DMA_RX_AGG_LIMIT_SHIFT) |
		/* Rx agg timeout in unit of 33ns */
		0x80;

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_USB_DMA_CFG, tmp);

	/* start up the receive pipe */

	for (i = 0; i < RT2870_SOFTC_RX_RING_DATA_COUNT; i++)
	{
		data = &sc->rx_ring.data[i];

		usbd_setup_xfer(data->xfer, sc->rx_ring.usb_pipe, data, data->buf,
			RT2870_USB_RX_BULK_BUFLEN, USBD_SHORT_XFER_OK | USBD_NO_COPY,
			USBD_NO_TIMEOUT, rt2870_rx_intr);

		usbd_transfer(data->xfer);
	}

	/* set Rx filter */

	tmp = RT2870_REG_RX_FILTER_DROP_CRC_ERR |
		RT2870_REG_RX_FILTER_DROP_PHY_ERR;

	if (ic->ic_opmode != IEEE80211_M_MONITOR)
	{
		tmp |= RT2870_REG_RX_FILTER_DROP_DUPL |
			RT2870_REG_RX_FILTER_DROP_CTS |
			RT2870_REG_RX_FILTER_DROP_BA |
			RT2870_REG_RX_FILTER_DROP_ACK |
			RT2870_REG_RX_FILTER_DROP_VER_ERR |
			RT2870_REG_RX_FILTER_DROP_CTRL_RSV |
			RT2870_REG_RX_FILTER_DROP_CFACK |
			RT2870_REG_RX_FILTER_DROP_CFEND;

		if (ic->ic_opmode == IEEE80211_M_STA)
			tmp |= RT2870_REG_RX_FILTER_DROP_RTS |
				RT2870_REG_RX_FILTER_DROP_PSPOLL;

		if (!(ifp->if_flags & IFF_PROMISC))
			tmp |= RT2870_REG_RX_FILTER_DROP_UC_NOME;
	}

	rt2870_io_mac_write(sc, RT2870_REG_RX_FILTER_CFG, tmp);

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL,
		RT2870_REG_RX_ENABLE | RT2870_REG_TX_ENABLE);

	/* clear garbage interrupts */

	tmp = rt2870_io_mac_read(sc, 0x1300);

	sc->cmd_ring.queued = 0;
	sc->cmd_ring.cur = 0;
	sc->cmd_ring.next = 0;

	ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
	ifp->if_drv_flags |= IFF_DRV_RUNNING;

	/*
	callout_reset(&sc->adjust_txpow_ch, hz, rt2870_adjust_txpow, sc);
	*/

	if (ic->ic_opmode != IEEE80211_M_MONITOR)
	{
		if (ic->ic_roaming != IEEE80211_ROAMING_MANUAL)
			ieee80211_new_state(ic, IEEE80211_S_SCAN, -1);
	}
	else
	{
		ieee80211_new_state(ic, IEEE80211_S_RUN, -1);
	}

	return;

fail:

	rt2870_stop(sc);
}

/*
 * rt2870_init_bbp
 */
static int rt2870_init_bbp(struct rt2870_softc *sc)
{
	int ntries, i;
	uint8_t tmp;

	for (ntries = 0; ntries < 20; ntries++)
	{
		tmp = rt2870_io_bbp_read(sc, 0);
		if (tmp != 0x00 && tmp != 0xff)
			break;
	}

	if (tmp == 0x00 || tmp == 0xff)
	{
		printf("%s: timeout waiting for BBP to wakeup\n",
			device_get_nameunit(sc->dev));
		return ETIMEDOUT;
	}

	for (i = 0; i < RT2870_DEF_BBP_SIZE; i++)
		rt2870_io_bbp_write(sc, rt2870_def_bbp[i].reg,
			rt2870_def_bbp[i].val);

	if ((sc->mac_rev & 0xffff) != 0x0101)
		rt2870_io_bbp_write(sc, 84, 0x19);

	if (sc->mac_rev == 0x28600100)
	{
		rt2870_io_bbp_write(sc, 69, 0x16);
		rt2870_io_bbp_write(sc, 73, 0x12);
	}

	return 0;
}

/*
 * rt2870_stop
 */
static void rt2870_stop(void *priv)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	uint32_t tmp;
	int i;

	sc = priv;
	ic = &sc->ic;
	ifp = ic->ic_ifp;

	if (ifp->if_drv_flags & IFF_DRV_RUNNING)
		rt2870_led_cmd(sc, RT2870_LED_CMD_RADIO_OFF);

	ifp->if_drv_flags &= ~(IFF_DRV_RUNNING | IFF_DRV_OACTIVE);

	ieee80211_new_state(ic, IEEE80211_S_INIT, -1);

	/* wait while cmd task is busy */

	while (sc->cmd_ring.queued > 0)
		tsleep(&sc->cmd_ring, 0, "cmd_ring", 0);

	usb_rem_task(sc->usb_dev, &sc->cmd_task);

	callout_stop(&sc->amrr_ch);
	callout_stop(&sc->tx_watchdog_ch);
	callout_stop(&sc->adjust_txpow_ch);

	/* clear key tables */

	rt2870_asic_clear_keytables(sc);

	/* disable Tx/Rx */

	tmp = rt2870_io_mac_read(sc, RT2870_REG_SYS_CTRL);

	tmp &= ~(RT2870_REG_RX_ENABLE | RT2870_REG_TX_ENABLE);

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL, tmp);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_USB_DMA_CFG, 0);

	/* reset adapter */

	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL,
		RT2870_REG_MAC_SRST | RT2870_REG_BBP_HRST);
	rt2870_io_mac_write(sc, RT2870_REG_SYS_CTRL, 0);

	if (sc->rx_ring.usb_pipe != NULL)
		usbd_abort_pipe(sc->rx_ring.usb_pipe);

	for (i = 0; i < RT2870_SOFTC_TX_RING_COUNT; i++)
		if (sc->tx_ring[i].usb_pipe != NULL)
			usbd_abort_pipe(sc->tx_ring[i].usb_pipe);
}

/*
 * rt2870_start
 */
static void rt2870_start(struct ifnet *ifp)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct ieee80211_node *ni;
	struct ether_header *eh;
	struct mbuf *m;
	int qid;

	sc = ifp->if_softc;
	ic = &sc->ic;

	for (;;)
	{
		IF_POLL(&ic->ic_mgtq, m);
		if (m != NULL)
		{
			if (sc->tx_ring[sc->tx_ring_mgtqid].queued >= RT2870_SOFTC_TX_RING_DATA_COUNT)
			{
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			IF_DEQUEUE(&ic->ic_mgtq, m);

			ni = (struct ieee80211_node *) m->m_pkthdr.rcvif;
			m->m_pkthdr.rcvif = NULL;

			if (bpf_peers_present(ic->ic_rawbpf))
				bpf_mtap(ic->ic_rawbpf, m);

			if (rt2870_tx_frame(sc, m, ni, sc->tx_ring_mgtqid) != 0)
				break;
		}
		else
		{
			if (ic->ic_state != IEEE80211_S_RUN)
				break;

			IF_POLL(&ifp->if_snd, m);
			if (m == NULL)
				break;

			qid = M_WME_GETAC(m);

			if (sc->tx_ring[qid].queued >= RT2870_SOFTC_TX_RING_DATA_COUNT)
			{
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
				break;
			}

			IFQ_DRV_DEQUEUE(&ifp->if_snd, m);

			if (ic->ic_flags & IEEE80211_F_SCAN)
				ieee80211_cancel_scan(ic);

			if (m->m_len < sizeof(struct ether_header) &&
				!(m = m_pullup(m, sizeof (struct ether_header))))
				continue;

			eh = mtod(m, struct ether_header *);

			ni = ieee80211_find_txnode(ic, eh->ether_dhost);
			if (ni == NULL)
			{
				m_freem(m);
				continue;
			}

			BPF_MTAP(ifp, m);

			m = ieee80211_encap(ic, m, ni);
			if (m == NULL)
			{
				ieee80211_free_node(ni);
				continue;
			}

			if (bpf_peers_present(ic->ic_rawbpf))
				bpf_mtap(ic->ic_rawbpf, m);

			if (rt2870_tx_frame(sc, m, ni, qid) != 0)
			{
				ieee80211_free_node(ni);
				ifp->if_oerrors++;
				break;
			}
		}

		sc->tx_timer = 5;

		ic->ic_lastdata = ticks;

		callout_reset(&sc->tx_watchdog_ch, hz, rt2870_tx_watchdog, sc);
	}
}

/*
 * rt2870_ioctl
 */
static int rt2870_ioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	int error;

	sc = ifp->if_softc;
	ic = &sc->ic;

	error = 0;

	switch (cmd)
	{
		case SIOCSIFFLAGS:
			if (ifp->if_flags & IFF_UP)
			{
				if (ifp->if_drv_flags & IFF_DRV_RUNNING)
					rt2870_asic_update_promisc(sc);
				else
					rt2870_init(sc);
			}
			else
			{
				if (ifp->if_drv_flags & IFF_DRV_RUNNING)
					rt2870_stop(sc);
			}
		break;

		default:
			error = ieee80211_ioctl(ic, cmd, data);
	}

	if (error == ENETRESET)
	{
		if ((ifp->if_flags & IFF_UP) &&
			(ifp->if_drv_flags & IFF_DRV_RUNNING) &&
			(ic->ic_roaming != IEEE80211_ROAMING_MANUAL))
		{
			rt2870_stop(sc);
			rt2870_init(sc);
		}

		error = 0;
	}

	return error;
}

/*
 * rt2870_newstate
 */
static int rt2870_newstate(struct ieee80211com *ic,
	enum ieee80211_state nstate, int arg)
{
	struct rt2870_softc *sc;
	struct rt2870_cmd_argv_newstate cmd_argv;

	sc = ic->ic_ifp->if_softc;

	callout_stop(&sc->amrr_ch);

	cmd_argv.state = nstate;
	cmd_argv.arg = arg;

	if (nstate == IEEE80211_S_INIT)
		sc->newstate(ic, nstate, arg);
	else
		rt2870_do_async(sc, rt2870_newstate_cb, &cmd_argv,
			sizeof(struct rt2870_cmd_argv_newstate));

	return 0;
}

/*
 * rt2870_scan_start
 */
static void rt2870_scan_start(struct ieee80211com *ic)
{
	struct rt2870_softc *sc;

	sc = ic->ic_ifp->if_softc;

	rt2870_do_async(sc, rt2870_scan_start_cb, NULL, 0);
}

/*
 * rt2870_scan_end
 */
static void rt2870_scan_end(struct ieee80211com *ic)
{
	struct rt2870_softc *sc;

	sc = ic->ic_ifp->if_softc;

	rt2870_do_async(sc, rt2870_scan_end_cb, NULL, 0);
}

/*
 * rt2870_set_channel
 */
static void rt2870_set_channel(struct ieee80211com *ic)
{
	struct rt2870_softc *sc;

	sc = ic->ic_ifp->if_softc;

	rt2870_do_async(sc, rt2870_set_channel_cb, NULL, 0);
}

/*
 * rt2870_newassoc
 */
static void rt2870_newassoc(struct ieee80211_node *ni, int isnew)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct rt2870_cmd_argv_newassoc cmd_argv;

	ic = ni->ni_ic;
	sc = ic->ic_ifp->if_softc;

	cmd_argv.associd = (ni != NULL) ? ni->ni_associd : 0;
	cmd_argv.isnew = isnew;
	memcpy(cmd_argv.macaddr, ni->ni_macaddr, IEEE80211_ADDR_LEN);

	rt2870_do_async(sc, rt2870_newassoc_cb, &cmd_argv,
		sizeof(struct rt2870_cmd_argv_newassoc));
}

/*
 * rt2870_updateslot
 */
static void rt2870_updateslot(struct ifnet *ifp)
{
	struct rt2870_softc *sc;

	sc = ifp->if_softc;

	rt2870_do_async(sc, rt2870_updateslot_cb, NULL, 0);
}

/*
 * rt2870_wme_update
 */
static int rt2870_wme_update(struct ieee80211com *ic)
{
	struct rt2870_softc *sc;

	sc = ic->ic_ifp->if_softc;

	rt2870_do_async(sc, rt2870_wme_update_cb, NULL, 0);

	return 0;
}

/*
 * rt2870_update_beacon
 */
static void rt2870_update_beacon(struct ieee80211com *ic, int what)
{
	struct rt2870_softc *sc;
	struct rt2870_cmd_argv_updatebeacon cmd_argv;

	sc = ic->ic_ifp->if_softc;

	cmd_argv.what = what;

	rt2870_do_async(sc, rt2870_update_beacon_cb, &cmd_argv,
		sizeof(struct rt2870_cmd_argv_updatebeacon));
}

/*
 * rt2870_key_set
 */
static int rt2870_key_set(struct ieee80211com *ic,
	const struct ieee80211_key *k, const uint8_t mac[IEEE80211_ADDR_LEN])
{
	struct rt2870_softc *sc;
	struct ieee80211_node *ni;
	struct rt2870_cmd_argv_keyset cmd_argv;

	if (k->wk_cipher->ic_cipher != IEEE80211_CIPHER_WEP &&
		k->wk_cipher->ic_cipher != IEEE80211_CIPHER_TKIP &&
		k->wk_cipher->ic_cipher != IEEE80211_CIPHER_AES_CCM)
		return EINVAL;

	sc = ic->ic_ifp->if_softc;
	ni = ic->ic_bss;

	memcpy(&cmd_argv.key, k, sizeof(struct ieee80211_key));
	cmd_argv.associd = (ni != NULL) ? ni->ni_associd : 0;

	rt2870_do_async(sc, rt2870_key_set_cb, &cmd_argv,
		sizeof(struct rt2870_cmd_argv_keyset));

	return 1;
}

/*
 * rt2870_key_delete
 */
static int rt2870_key_delete(struct ieee80211com *ic,
	const struct ieee80211_key *k)
{
	struct rt2870_softc *sc;
	struct ieee80211_node *ni;
	struct rt2870_cmd_argv_keydelete cmd_argv;

	sc = ic->ic_ifp->if_softc;
	ni = ic->ic_bss;

	memcpy(&cmd_argv.key, k, sizeof(struct ieee80211_key));
	cmd_argv.associd = (ni != NULL) ? ni->ni_associd : 1;

	rt2870_do_async(sc, rt2870_key_delete_cb, &cmd_argv,
		sizeof(struct rt2870_cmd_argv_keydelete));

	return 1;
}

/*
 * rt2870_raw_xmit
 */
static int rt2870_raw_xmit(struct ieee80211_node *ni, struct mbuf *m,
	const struct ieee80211_bpf_params *params)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct rt2870_softc *sc;

	ic = ni->ni_ic;
	ifp = ic->ic_ifp;
	sc = ifp->if_softc;

	return 0;
}

/*
 * rt2870_media_change
 */
static int rt2870_media_change(struct ifnet *ifp)
{
	struct rt2870_softc *sc;
	int error;

	sc = ifp->if_softc;

	error = ieee80211_media_change(ifp);
	if (error != ENETRESET)
		return error;

	if ((ifp->if_flags & IFF_UP) && (ifp->if_drv_flags & IFF_DRV_RUNNING))
	{
		rt2870_stop(sc);
		rt2870_init(sc);
	}

	return 0;
}

/*
 * rt2870_amrr_start
 */
static void rt2870_amrr_start(struct rt2870_softc *sc,
	struct ieee80211_node *ni)
{
	int i;

	ieee80211_amrr_node_init(&sc->amrr, &sc->amrr_node);

	for (i = ni->ni_rates.rs_nrates - 1;
		i > 0 && (ni->ni_rates.rs_rates[i] & IEEE80211_RATE_VAL) > 72;
		i--);

	ni->ni_txrate = i;

	callout_reset(&sc->amrr_ch, hz, rt2870_amrr_timeout, sc);
}

/*
 * rt2870_amrr_timeout
 */
static void rt2870_amrr_timeout(void *arg)
{
	struct rt2870_softc *sc;
	usb_device_request_t req;

	sc = arg;

	req.bmRequestType = UT_READ_VENDOR_DEVICE;
	req.bRequest = RT2870_USB_REQ_MAC_READ_MULTI;
	USETW(req.wValue, 0);
	USETW(req.wIndex, RT2870_REG_RX_STA_CNT0);
	USETW(req.wLength, sizeof(sc->amrr_sta));

	usbd_setup_default_xfer(sc->amrr_xfer, sc->usb_dev, sc,
		USBD_DEFAULT_TIMEOUT, &req, sc->amrr_sta, sizeof(sc->amrr_sta),
		0, rt2870_amrr_update);

	usbd_transfer(sc->amrr_xfer);
}

/*
 * rt2870_amrr_update
 */
static void rt2870_amrr_update(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	struct ifnet *ifp;
	uint32_t failcnt, txcnt, retrycnt;

	sc = priv;
	ic = &sc->ic;
	ifp = sc->ic.ic_ifp;

	if (status != USBD_NORMAL_COMPLETION)
	{
		if (status == USBD_NOT_STARTED || status == USBD_CANCELLED)
			return;

		printf("%s: could not retrieve Tx statistics\n",
			device_get_nameunit(sc->dev));

		goto skip;
	}

	failcnt = (le32toh(sc->amrr_sta[0]) & 0xffff);
	txcnt = (le32toh(sc->amrr_sta[1]) & 0xffff);
	retrycnt = (le32toh(sc->amrr_sta[1]) >> 16);

	DPRINTFN(sc, 3, ("%s: Tx statistics: failcnt=%d, txcnt=%d, retrycnt=%d\n",
		device_get_nameunit(sc->dev),
		failcnt, txcnt, retrycnt));

	ifp->if_oerrors += failcnt;

	sc->amrr_node.amn_retrycnt = failcnt + retrycnt;
	sc->amrr_node.amn_txcnt = failcnt + retrycnt + txcnt;

	ieee80211_amrr_choose(&sc->amrr, sc->ic.ic_bss, &sc->amrr_node);

skip:

	callout_reset(&sc->amrr_ch, hz, rt2870_amrr_timeout, sc);
}

/*
 * rt2870_tx_watchdog
 */
static void rt2870_tx_watchdog(void *arg)
{
	struct rt2870_softc *sc;

	sc = arg;

	if (sc->tx_timer > 0)
	{
		if (--sc->tx_timer == 0)
		{
			printf("%s: device timeout\n",
				device_get_nameunit(sc->dev));
			sc->ifp->if_oerrors++;
			return;
		}

		callout_reset(&sc->tx_watchdog_ch, hz, rt2870_tx_watchdog, sc);
	}
}

/*
 * rt2870_adjust_txpow
 */
static void rt2870_adjust_txpow(void *arg)
{
	struct rt2870_softc *sc;
	struct ieee80211com *ic;
	int8_t delta_pow;
	int tx_agc_cntl, i;
	const uint8_t *tssi;
	uint8_t tssi_step;
	uint8_t temperature;

	sc = arg;
	ic = &sc->ic;

	delta_pow = 0;

	if (IEEE80211_IS_CHAN_2GHZ(ic->ic_bss->ni_chan))
	{
		tx_agc_cntl = sc->tx_agc_cntl_2ghz;
		tssi = &sc->tssi_2ghz[4];
		tssi_step = sc->tssi_step_2ghz;
	} else
	{
		tx_agc_cntl = sc->tx_agc_cntl_5ghz;
		tssi = &sc->tssi_5ghz[4];
		tssi_step = sc->tssi_step_5ghz;
	}

	if (tx_agc_cntl)
	{
		temperature = rt2870_io_bbp_read(sc, 49);

		if (temperature > tssi[1])
		{
			for (i = 1; i < 5; i++)
			{
				if (temperature <= tssi[i])
					break;
			}
		}
		else if (temperature < tssi[1])
		{
			for (i = 1; i < 5; i++)
			{
				if (temperature >= tssi[i])
					break;
			}
		}
		else
		{
		}
	}

	callout_reset(&sc->adjust_txpow_ch, hz, rt2870_adjust_txpow, sc);
}

/*
 * rt2870_newstate_cb
 */
static void rt2870_newstate_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	enum ieee80211_state ostate;
	struct ieee80211_node *ni;
	struct rt2870_cmd_argv_newstate *cmd_argv;

	ic = &sc->ic;
	cmd_argv = arg;

	ostate = ic->ic_state;

	/* turn link LED off */

	if (ostate == IEEE80211_S_RUN)
		rt2870_led_cmd(sc, RT2870_LED_CMD_RADIO_OFF);

	switch (cmd_argv->state)
	{
		case IEEE80211_S_INIT:
			rt2870_asic_disable_tsf_sync(sc);
		break;

		case IEEE80211_S_RUN:
			ni = ic->ic_bss;

			if (ic->ic_opmode != IEEE80211_M_MONITOR)
			{
				rt2870_asic_enable_mrr(sc);
				rt2870_asic_set_txpreamble(sc);
				rt2870_asic_set_basicrates(sc);
				rt2870_asic_set_bssid(sc, ni->ni_bssid);
			}

			if (ic->ic_opmode == IEEE80211_M_STA)
				rt2870_newassoc(ni, 1);
			
			if (ic->ic_opmode != IEEE80211_M_MONITOR)
				rt2870_asic_enable_tsf_sync(sc);

			/* turn link LED on */

			if (ic->ic_opmode != IEEE80211_M_MONITOR)
			{
				rt2870_led_cmd(sc, RT2870_LED_CMD_RADIO_ON |
		    		(IEEE80211_IS_CHAN_2GHZ(ni->ni_chan) ?
		     			RT2870_LED_CMD_LINK_2GHZ : RT2870_LED_CMD_LINK_5GHZ));
			}
		break;

		default:
		break;
	}

	sc->newstate(ic, cmd_argv->state, cmd_argv->arg);
}

/*
 * rt2870_scan_start_cb
 */
static void rt2870_scan_start_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;

	ic = &sc->ic;
	ifp = ic->ic_ifp;

	rt2870_asic_disable_tsf_sync(sc);
}

/*
 * rt2870_scan_end_cb
 */
static void rt2870_scan_end_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	struct ieee80211_node *ni;

	ic = &sc->ic;
	ni = ic->ic_bss;

	rt2870_asic_enable_tsf_sync(sc);
}

/*
 * rt2870_set_channel_cb
 */
static void rt2870_set_channel_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;

	ic = &sc->ic;

	rt2870_rf_set_chan(sc, ic->ic_curchan);
}

/*
 * rt2870_newassoc_cb
 */
static void rt2870_newassoc_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	struct ieee80211_node *ni;
	struct rt2870_cmd_argv_newassoc *cmd_argv;
	uint16_t associd;
	int isnew;
	uint8_t *macaddr, wcid;

	ic = &sc->ic;
	ni = ic->ic_bss;
	cmd_argv = arg;
	associd = cmd_argv->associd;
	isnew = cmd_argv->isnew;
	macaddr = cmd_argv->macaddr;

	wcid = RT2870_AID2WCID(associd);

	DPRINTFN(sc, 3, ("%s: new association: wcid=0x%02x, "
		"mac addr=%02x:%02x:%02x:%02x:%02x:%02x\n",
		device_get_nameunit(sc->dev), wcid,
		macaddr[0], macaddr[1], macaddr[2],
		macaddr[3], macaddr[4], macaddr[5]));

	rt2870_io_mac_write_multi(sc, RT2870_REG_WCID(wcid),
 		macaddr, IEEE80211_ADDR_LEN);

	if (ic->ic_fixed_rate == IEEE80211_FIXED_RATE_NONE)
		rt2870_amrr_start(sc, ni);
}

/*
 * rt2870_updateslot_cb
 */
static void rt2870_updateslot_cb(struct rt2870_softc *sc, void *arg)
{
	rt2870_asic_updateslot(sc);
}

/*
 * rt2870_wme_update_cb
 */
static void rt2870_wme_update_cb(struct rt2870_softc *sc, void *arg)
{
	rt2870_asic_wme_update(sc);
}

/*
 * rt2870_update_beacon_cb
 */
static void rt2870_update_beacon_cb(struct rt2870_softc *sc, void *arg)
{
	struct rt2870_cmd_argv_updatebeacon *cmd_argv;

	cmd_argv = arg;

	rt2870_asic_update_beacon(sc);
}

/*
 * rt2870_key_set_cb
 */
static void rt2870_key_set_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	struct ieee80211_key *k;
	struct rt2870_cmd_argv_keyset *cmd_argv;
	uint16_t associd, key_base, keymode_base;
	uint8_t mode, vap, wcid, iv[8];
	uint32_t tmp;

	ic = &sc->ic;
	cmd_argv = arg;
	k = &cmd_argv->key;
	associd = cmd_argv->associd;

	switch (k->wk_cipher->ic_cipher)
	{
		case IEEE80211_CIPHER_WEP:
			if(k->wk_keylen < 8)
				mode = RT2870_REG_CIPHER_MODE_WEP40;
			else
				mode = RT2870_REG_CIPHER_MODE_WEP104;
		break;

		case IEEE80211_CIPHER_TKIP:
			mode = RT2870_REG_CIPHER_MODE_TKIP;
		break;

		case IEEE80211_CIPHER_AES_CCM:
			mode = RT2870_REG_CIPHER_MODE_AES_CCMP;
		break;

		default:
			return;
	}

	DPRINTFN(sc, 3, ("%s: set key: keyix=%d, keylen=%d, associd=0x%04x, mode=%d, group=%d\n",
		device_get_nameunit(sc->dev), k->wk_keyix, k->wk_keylen, associd, mode,
		(k->wk_flags & IEEE80211_KEY_GROUP) ? 1 : 0));

	if (!(k->wk_flags & IEEE80211_KEY_GROUP))
	{
		/* install pairwise key */

		vap = 0;
		wcid = RT2870_AID2WCID(associd);
		key_base = RT2870_REG_PKEY(wcid);

		if (k->wk_cipher->ic_cipher == IEEE80211_CIPHER_WEP)
		{
			memset(iv, 0, 8);

			iv[3] = (k->wk_keyix << 6);
		}
		else
		{
			if (k->wk_cipher->ic_cipher == IEEE80211_CIPHER_TKIP)
			{
				iv[0] = (k->wk_keytsc >> 8);
				iv[1] = ((iv[0] | 0x20) & 0x7f);
				iv[2] = k->wk_keytsc;
			}
			else
			{
				/* AES CCMP */

				iv[0] = k->wk_keytsc;
				iv[1] = k->wk_keytsc >> 8;
				iv[2] = 0;
			}

			iv[3] = ((k->wk_keyix << 6) | IEEE80211_WEP_EXTIV);
			iv[4] = (k->wk_keytsc >> 16);
			iv[5] = (k->wk_keytsc >> 24);
			iv[6] = (k->wk_keytsc >> 32);
			iv[7] = (k->wk_keytsc >> 40);

			DPRINTFN(sc, 3, ("%s: set key: iv=%02x %02x %02x %02x %02x %02x %02x %02x\n",
				device_get_nameunit(sc->dev),
				iv[0], iv[1], iv[2], iv[3], iv[4], iv[5], iv[6], iv[7]));
		}

		rt2870_io_mac_write_multi(sc, RT2870_REG_IVEIV(wcid), iv, 8);

		if (k->wk_cipher->ic_cipher == IEEE80211_CIPHER_TKIP)
		{
			rt2870_io_mac_write_multi(sc, key_base, k->wk_key, 16);

			if (ic->ic_opmode != IEEE80211_M_HOSTAP)
			{
				rt2870_io_mac_write_multi(sc, key_base + 16, &k->wk_key[16], 8);
				rt2870_io_mac_write_multi(sc, key_base + 24, &k->wk_key[24], 8);
			}
			else
			{
				rt2870_io_mac_write_multi(sc, key_base + 16, &k->wk_key[24], 8);
				rt2870_io_mac_write_multi(sc, key_base + 24, &k->wk_key[16], 8);
			}
		}
		else
		{
			rt2870_io_mac_write_multi(sc, key_base, k->wk_key, k->wk_keylen);
		}

		tmp = ((vap & RT2870_REG_VAP_MASK) << RT2870_REG_VAP_SHIFT) |
			(mode << RT2870_REG_CIPHER_MODE_SHIFT) | RT2870_REG_PKEY_ENABLE;

		rt2870_io_mac_write(sc, RT2870_REG_WCID_ATTR(wcid), tmp);
	}

	if ((k->wk_flags & IEEE80211_KEY_GROUP) ||
		(k->wk_cipher->ic_cipher == IEEE80211_CIPHER_WEP))
	{
		/* install group key */

		vap = 0;
		key_base = RT2870_REG_SKEY(vap, k->wk_keyix);
		keymode_base = RT2870_REG_SKEY_MODE(vap);

		if (k->wk_cipher->ic_cipher == IEEE80211_CIPHER_TKIP)
		{
			rt2870_io_mac_write_multi(sc, key_base, k->wk_key, 16);

			if (ic->ic_opmode != IEEE80211_M_HOSTAP)
			{
				rt2870_io_mac_write_multi(sc, key_base + 16, &k->wk_key[16], 8);
				rt2870_io_mac_write_multi(sc, key_base + 24, &k->wk_key[24], 8);
			}
			else
			{
				rt2870_io_mac_write_multi(sc, key_base + 16, &k->wk_key[24], 8);
				rt2870_io_mac_write_multi(sc, key_base + 24, &k->wk_key[16], 8);
			}
		}
		else
		{
			rt2870_io_mac_write_multi(sc, key_base, k->wk_key, k->wk_keylen);
		}

		tmp = rt2870_io_mac_read(sc, keymode_base);

		tmp &= ~(0xf << (k->wk_keyix * 4 + 16 * (vap % 2)));
		tmp |= (mode << (k->wk_keyix * 4 + 16 * (vap % 2)));

		rt2870_io_mac_write(sc, keymode_base, tmp);
	}

}

/*
 * rt2870_key_delete_cb
 */
static void rt2870_key_delete_cb(struct rt2870_softc *sc, void *arg)
{
	struct ieee80211com *ic;
	struct ieee80211_key *k;
	struct rt2870_cmd_argv_keydelete *cmd_argv;
	uint16_t associd;

	ic = &sc->ic;
	cmd_argv = arg;
	k = &cmd_argv->key;
	associd = cmd_argv->associd;
}

/*
 * rt2870_do_async
 */
static int rt2870_do_async(struct rt2870_softc *sc,
	void (*cb)(struct rt2870_softc *sc, void *arg),
	void *arg, int len)
{
	struct rt2870_softc_cmd_ring *ring;
	struct rt2870_softc_cmd *cmd;

	ring = &sc->cmd_ring;

	if (ring->queued >= RT2870_SOFTC_CMD_RING_CMD_COUNT)
		return -1;

	ring->queued++;

	cmd = &ring->cmd[ring->cur];

	cmd->cb = cb;

	ring->cur = (ring->cur + 1) % RT2870_SOFTC_CMD_RING_CMD_COUNT;

	if(arg != NULL)
		memcpy(cmd->data, arg, len);

	if (ring->queued == 1)
		usb_add_task(sc->usb_dev, &sc->cmd_task, USB_TASKQ_DRIVER);

	return 0;
}

/*
 * rt2870_cmd_task
 */
static void rt2870_cmd_task(void *arg)
{
	struct rt2870_softc *sc;
	struct rt2870_softc_cmd_ring *ring;
	struct rt2870_softc_cmd *cmd;

	sc = arg;
	ring = &sc->cmd_ring;

	while (ring->next != ring->cur)
	{
		cmd = &ring->cmd[ring->next];

		cmd->cb(sc, cmd->data);

		ring->queued--;
		ring->next = (ring->next + 1) % RT2870_SOFTC_CMD_RING_CMD_COUNT;
	}

	wakeup(ring);
}

/*
 * rt2870_asic_set_bssid
 */
static void rt2870_asic_set_bssid(struct rt2870_softc *sc,
	const uint8_t *bssid)
{
	uint32_t tmp;

	DPRINTFN(sc, 3, ("%s: set bssid: bssid=%02x:%02x:%02x:%02x:%02x:%02x\n",
		device_get_nameunit(sc->dev),
		bssid[0], bssid[1], bssid[2], bssid[3],
		bssid[4], bssid[5]));

	tmp = bssid[0] | (bssid[1]) << 8 | (bssid[2] << 16) | (bssid[3] << 24);

	rt2870_io_mac_write(sc, RT2870_REG_BSSID_DW0, tmp);

	tmp = bssid[4] | (bssid[5] << 8);

	rt2870_io_mac_write(sc, RT2870_REG_BSSID_DW1, tmp);
}

/*
 * rt2870_asic_set_macaddr
 */
static void rt2870_asic_set_macaddr(struct rt2870_softc *sc,
	const uint8_t *addr)
{
	uint32_t tmp;

	tmp = addr[0] | (addr[1] << 8) | (addr[2] << 16) | (addr[3] << 24);

	rt2870_io_mac_write(sc, RT2870_REG_ADDR_DW0, tmp);

	tmp = addr[4] | (addr[5] << 8);

	rt2870_io_mac_write(sc, RT2870_REG_ADDR_DW1, tmp);
}

/*
 * rt2870_asic_enable_tsf_sync
 */
static void rt2870_asic_enable_tsf_sync(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	uint32_t tmp;

	ic = &sc->ic;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_BCN_TIME_CFG);

	tmp &= ~0x1fffff;
	tmp |= ic->ic_bss->ni_intval * 16;
	tmp |= (RT2870_REG_TSF_TIMER_ENABLE | RT2870_REG_TBTT_TIMER_ENABLE);

	if (ic->ic_opmode == IEEE80211_M_STA)
	{
		tmp |= (RT2870_REG_TSF_SYNC_MODE_STA << RT2870_REG_TSF_SYNC_MODE_SHIFT);
	}
	else if (ic->ic_opmode == IEEE80211_M_IBSS)
	{
		tmp |= RT2870_REG_BCN_TX_ENABLE;
		tmp |= (RT2870_REG_TSF_SYNC_MODE_IBSS << RT2870_REG_TSF_SYNC_MODE_SHIFT);
	}
	else if (ic->ic_opmode == IEEE80211_M_HOSTAP)
	{
		tmp |= RT2870_REG_BCN_TX_ENABLE;
		tmp |= (RT2870_REG_TSF_SYNC_MODE_HOSTAP << RT2870_REG_TSF_SYNC_MODE_SHIFT);
	}

	rt2870_io_mac_write(sc, RT2870_REG_BCN_TIME_CFG, tmp);
}

/*
 * rt2870_asic_disable_tsf_sync
 */
static void rt2870_asic_disable_tsf_sync(struct rt2870_softc *sc)
{
	uint32_t tmp;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_BCN_TIME_CFG);

	tmp &= ~(RT2870_REG_BCN_TX_ENABLE |
		RT2870_REG_TSF_TIMER_ENABLE |
		RT2870_REG_TBTT_TIMER_ENABLE);

	tmp &= ~(RT2870_REG_TSF_SYNC_MODE_MASK << RT2870_REG_TSF_SYNC_MODE_SHIFT);
	tmp |= (RT2870_REG_TSF_SYNC_MODE_DISABLE << RT2870_REG_TSF_SYNC_MODE_SHIFT);

	rt2870_io_mac_write(sc, RT2870_REG_BCN_TIME_CFG, tmp);
}

/*
 * rt2870_asic_update_promisc
 */
static void rt2870_asic_update_promisc(struct rt2870_softc *sc)
{
	struct ifnet *ifp;
	uint32_t tmp;

	ifp = sc->ic.ic_ifp;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_RX_FILTER_CFG);

	tmp &= ~RT2870_REG_RX_FILTER_DROP_UC_NOME;

	if (!(ifp->if_flags & IFF_PROMISC))
		tmp |= RT2870_REG_RX_FILTER_DROP_UC_NOME;

	rt2870_io_mac_write(sc, RT2870_REG_RX_FILTER_CFG, tmp);

	printf("%s: %s promiscuous mode\n",
		device_get_nameunit(sc->dev),
		(ifp->if_flags & IFF_PROMISC) ? "entering" : "leaving");
}

/*
 * rt2870_asic_enable_mrr
 */
static void rt2870_asic_enable_mrr(struct rt2870_softc *sc)
{
#define CCK(mcs)	(mcs)
#define OFDM(mcs)	((1 << 3) | (mcs))

	rt2870_io_mac_write(sc, RT2870_REG_TX_LG_FBK_CFG0,
		OFDM(6) << 28 |	/* 54 -> 48 */
		OFDM(5) << 24 |	/* 48 -> 36 */
		OFDM(4) << 20 |	/* 36 -> 24 */
		OFDM(3) << 16 |	/* 24 -> 18 */
		OFDM(2) << 12 |	/* 18 -> 12 */
		OFDM(1) << 8  |	/* 12 -> 9 */
		OFDM(0) << 4  |	/*  9 -> 6 */
		OFDM(0));		/*  6 -> 6 */

	rt2870_io_mac_write(sc, RT2870_REG_TX_LG_FBK_CFG1,
		CCK(2) << 12 |	/* 11  -> 5.5 */
		CCK(1) << 8  |	/* 5.5 -> 2 */
		CCK(0) << 4  |	/*   2 -> 1 */
		CCK(0));		/*   1 -> 1 */

#undef OFDM
#undef CCK
}

/*
 * rt2870_asic_set_txpreamble
 */
static void rt2870_asic_set_txpreamble(struct rt2870_softc *sc)
{
	uint32_t tmp;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_AUTO_RSP_CFG);

	tmp &= ~RT2870_REG_CCK_SHORT_ENABLE;

	if (sc->ic.ic_flags & IEEE80211_F_SHPREAMBLE)
		tmp |= RT2870_REG_CCK_SHORT_ENABLE;

	rt2870_io_mac_write(sc, RT2870_REG_AUTO_RSP_CFG, tmp);
}

/*
 * rt2870_asic_set_basicrates
 */
static void rt2870_asic_set_basicrates(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;

	ic = &sc->ic;

	if (ic->ic_curmode == IEEE80211_MODE_11B)
		rt2870_io_mac_write(sc, RT2870_REG_LEGACY_BASIC_RATE, 0x3);
	else if (ic->ic_curmode == IEEE80211_MODE_11A)
		rt2870_io_mac_write(sc, RT2870_REG_LEGACY_BASIC_RATE, 0x150);
	else
		rt2870_io_mac_write(sc, RT2870_REG_LEGACY_BASIC_RATE, 0x15f);
}

/*
 * rt2870_asic_update_rtsthreshold
 */
static void rt2870_asic_update_rtsthreshold(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	uint32_t tmp;
	uint16_t threshold;

	ic = &sc->ic;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_TX_RTS_CFG);

	tmp &= ~(RT2870_REG_TX_RTS_THRESHOLD_MASK << RT2870_REG_TX_RTS_THRESHOLD_SHIFT);

	threshold = (ic->ic_rtsthreshold < IEEE80211_RTS_MAX) ?
		ic->ic_rtsthreshold : 0x1000;

	tmp |= ((threshold & RT2870_REG_TX_RTS_THRESHOLD_MASK) <<
		RT2870_REG_TX_RTS_THRESHOLD_SHIFT);

	rt2870_io_mac_write(sc, RT2870_REG_TX_RTS_CFG, tmp);
}

/*
 * rt2870_asic_updateprot
 */
static void rt2870_asic_updateprot(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	uint32_t tmp;

	ic = &sc->ic;

	tmp = RT2870_REG_RTSTH_ENABLE | RT2870_REG_PROT_NAV_SHORT |
		RT2870_REG_TXOP_ALLOW_ALL;

	/* set up protection frame rate (MCS code) */

	tmp |= (ic->ic_curmode == IEEE80211_MODE_11A) ? 0 : 3;

	/* CCK frames don't require protection */

	rt2870_io_mac_write(sc, RT2870_REG_TX_CCK_PROT_CFG, tmp);

	if (ic->ic_flags & IEEE80211_F_USEPROT)
	{
		if (ic->ic_protmode == IEEE80211_PROT_RTSCTS)
			tmp |= RT2870_REG_PROT_CTRL_RTS_CTS;
		else if (ic->ic_protmode == IEEE80211_PROT_CTSONLY)
			tmp |= RT2870_REG_PROT_CTRL_CTS;
	}

	rt2870_io_mac_write(sc, RT2870_REG_TX_OFDM_PROT_CFG, tmp);
}

/*
 * rt2870_asic_updateslot
 */
static void rt2870_asic_updateslot(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	uint32_t tmp;

	ic = &sc->ic;

	tmp = rt2870_io_mac_read(sc, RT2870_REG_BKOFF_SLOT_CFG);

	tmp &= ~0xff;
	tmp |= (ic->ic_flags & IEEE80211_F_SHSLOT) ? 9 : 20;

	rt2870_io_mac_write(sc, RT2870_REG_BKOFF_SLOT_CFG, tmp);
}

/*
 * rt2870_asic_wme_update
 */
static void rt2870_asic_wme_update(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	struct ieee80211_wme_state *wme;
	int i;

	ic = &sc->ic;
	wme = &ic->ic_wme;

	DPRINTF(sc, ("%s: wme update\n",
		device_get_nameunit(sc->dev)));

	for (i = 0; i < WME_NUM_AC; i++)
		rt2870_io_mac_write(sc, RT2870_REG_TX_EDCA_AC_CFG(i),
			(wme->wme_params[i].wmep_logcwmax << 16) |
			(wme->wme_params[i].wmep_logcwmin << 12) |
			(wme->wme_params[i].wmep_aifsn << 8) |
			wme->wme_params[i].wmep_txopLimit);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_AIFSN_CFG,
		(wme->wme_params[WME_AC_VO].wmep_aifsn << 12) |
		(wme->wme_params[WME_AC_VI].wmep_aifsn << 8) |
		(wme->wme_params[WME_AC_BK].wmep_aifsn << 4) |
		wme->wme_params[WME_AC_BE].wmep_aifsn);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_CWMIN_CFG,
		(wme->wme_params[WME_AC_VO].wmep_logcwmin << 12) |
		(wme->wme_params[WME_AC_VI].wmep_logcwmin << 8) |
		(wme->wme_params[WME_AC_BK].wmep_logcwmin << 4) |
		wme->wme_params[WME_AC_BE].wmep_logcwmin);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_CWMAX_CFG,
		(wme->wme_params[WME_AC_VO].wmep_logcwmax << 12) |
		(wme->wme_params[WME_AC_VI].wmep_logcwmax << 8) |
		(wme->wme_params[WME_AC_BK].wmep_logcwmax << 4) |
		wme->wme_params[WME_AC_BE].wmep_logcwmax);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_TXOP0_CFG,
		(wme->wme_params[WME_AC_BK].wmep_txopLimit << 16) |
		wme->wme_params[WME_AC_BE].wmep_txopLimit);

	rt2870_io_mac_write(sc, RT2870_REG_SCHDMA_WMM_TXOP1_CFG,
		(wme->wme_params[WME_AC_VO].wmep_txopLimit << 16) |
		wme->wme_params[WME_AC_VI].wmep_txopLimit);
}

/*
 * rt2870_asic_update_beacon
 */
static void rt2870_asic_update_beacon(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	struct mbuf *m;
	struct rt2870_txwi txwi;
	uint8_t rate, mcs;

	ic = &sc->ic;

	m = ieee80211_beacon_alloc(ic->ic_bss, &sc->beacon_offsets);
	if (m == NULL)
		return;

	rate = IEEE80211_IS_CHAN_5GHZ(ic->ic_curchan) ? 12 : 2;
	mcs = rt2870_rate2mcs(rate);

	memset(&txwi, 0 ,sizeof(struct rt2870_txwi));

	txwi.wcid = 0xff;
	txwi.pid_mpdu_len = ((htole16(m->m_pkthdr.len) & RT2870_TXWI_MPDU_LEN_MASK) <<
		 RT2870_TXWI_MPDU_LEN_SHIFT);
	txwi.txop = (RT2870_TXWI_TXOP_HT << RT2870_TXWI_TXOP_SHIFT);
	txwi.mpdu_density_flags |=
		(RT2870_TXWI_FLAGS_TS << RT2870_TXWI_FLAGS_SHIFT);

	if (rate == 2)
		txwi.phymode_ifs_stbc_shortgi =
			(RT2870_TXWI_PHYMODE_CCK << RT2870_TXWI_PHYMODE_SHIFT);
	else
		txwi.phymode_ifs_stbc_shortgi =
			(RT2870_TXWI_PHYMODE_OFDM << RT2870_TXWI_PHYMODE_SHIFT);

	txwi.bw_mcs = ((mcs & RT2870_TXWI_MCS_MASK) << RT2870_TXWI_MCS_SHIFT);

	rt2870_io_mac_write_multi(sc, RT2870_REG_BEACON_BASE(0),
		&txwi, sizeof(struct rt2870_txwi));

	rt2870_io_mac_write_multi(sc, RT2870_REG_BEACON_BASE(0) + sizeof(struct rt2870_txwi),
		mtod(m, uint8_t *), m->m_pkthdr.len);

	m_freem(m);
}

/*
 * rt2870_asic_clear_keytables
 */
static void rt2870_asic_clear_keytables(struct rt2870_softc *sc)
{
	/* clear Rx WCID search table (entries = 256, entry size = 8) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_WCID(0), 0, 2 * 256);

	/* clear WCID attribute table (entries = 256, entry size = 4) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_WCID_ATTR(0), 0, 256);

	/* clear IV/EIV table (entries = 256, entry size = 8) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_IVEIV(0), 0, 2 * 256);

	/* clear pairwise key table (entries = 256, entry size = 32) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_PKEY(0), 0, 8 * 256);

	/* clear shared key table (entries = 32, entry size = 32) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_SKEY(0, 0), 0, 8 * 32);

	/* clear shared key mode (entries = 32, entry size = 2) */

	rt2870_io_mac_set_region_4(sc, RT2870_REG_SKEY_MODE(0), 0, 16);
}

/*
 * rt2870_rxrate
 */
static uint8_t rt2870_rxrate(struct rt2870_rxwi *rxwi)
{
	uint8_t mcs, phymode;
	uint8_t rate;

	mcs = (rxwi->bw_mcs >> RT2870_RXWI_MCS_SHIFT) & RT2870_RXWI_MCS_MASK;
	phymode = (rxwi->phymode_stbc_shortgi >> RT2870_RXWI_PHYMODE_SHIFT) &
		RT2870_RXWI_PHYMODE_MASK;

	rate = 2;

	switch (phymode)
	{
		case RT2870_RXWI_PHYMODE_CCK:
			switch (mcs & ~RT2870_RXWI_MCS_SHOTPRE)
			{
				case 0: rate = 2; break;
				case 1: rate = 4; break;
				case 2: rate = 11; break;
				case 3: rate = 22; break;
			}
		break;

		case RT2870_RXWI_PHYMODE_OFDM:
			switch (mcs)
			{
				case 0: rate = 12; break;
				case 1: rate = 18; break;
				case 2: rate = 24; break;
				case 3: rate = 36; break;
				case 4: rate = 48; break;
				case 5: rate = 72; break;
				case 6: rate = 96; break;
				case 7: rate = 108; break;
			}
		break;
	}

	return rate;
}

/*
 * rt2870_maxrssi_rxpath
 */
static uint8_t rt2870_maxrssi_rxpath(struct rt2870_softc *sc,
	const struct rt2870_rxwi *rxwi)
{
	uint8_t rxpath;

	rxpath = 0;

	if (sc->nrxpath > 1)
		if (rxwi->rssi[1] > rxwi->rssi[rxpath])
			rxpath = 1;

	if (sc->nrxpath > 2)
		if (rxwi->rssi[2] > rxwi->rssi[rxpath])
			rxpath = 2;

	return rxpath;
}

/*
 * rt2870_rssi2dbm
 */
static int8_t rt2870_rssi2dbm(struct rt2870_softc *sc,
	uint8_t rssi, uint8_t rxpath)
{
	struct ieee80211com *ic;
	struct ieee80211_channel *c;
	int chan, delta;

	ic = &sc->ic;
	c = ic->ic_curchan;
	chan = ieee80211_chan2ieee(ic, c);

	if (IEEE80211_IS_CHAN_5GHZ(c))
	{
		delta = sc->rssi_off_5ghz[rxpath];

		if (chan <= 64)
			delta -= sc->lna_gain[1];
		else if (chan <= 128)
			delta -= sc->lna_gain[2];
		else
			delta -= sc->lna_gain[3];
	}
	else
	{
		delta = sc->rssi_off_2ghz[rxpath] - sc->lna_gain[0];
	}

	return (-12 - delta - rssi);
}

/*
 * rt2870_rate2mcs
 */
static uint8_t rt2870_rate2mcs(uint8_t rate)
{
	switch (rate)
	{
		/* CCK rates */
		case 2:	return 0;
		case 4:	return 1;
		case 11: return 2;
		case 22: return 3;

		/* OFDM rates */
		case 12: return 0;
		case 18: return 1;
		case 24: return 2;
		case 36: return 3;
		case 48: return 4;
		case 72: return 5;
		case 96: return 6;
		case 108: return 7;
	}

	return 0;
}

/*
 * rt2870_ackrate
 */
static int rt2870_ackrate(struct ieee80211com *ic, int rate)
{
	switch (rate)
	{
		/* CCK rates */

		case 2:
			return 2;

		case 4:
		case 11:
		case 22:
			return (ic->ic_curmode == IEEE80211_MODE_11B) ? 4 : rate;

		/* OFDM rates */

		case 12:
		case 18:
			return 12;

		case 24:
		case 36:
			return 24;

		case 48:
		case 72:
		case 96:
		case 108:
			return 48;
	}

	/* default to 1Mbps */
	return 2;
}

/*
 * rt2870_txtime
 */
static uint16_t rt2870_txtime(int len, int rate, uint32_t flags)
{
	uint16_t txtime;

	if (RT2870_RATE_IS_OFDM(rate))
	{
		txtime = (8 + 4 * len + 3 + rate - 1) / rate;
		txtime = 16 + 4 + 4 * txtime + 6;
	}
	else
	{
		txtime = (16 * len + rate - 1) / rate;

		if (rate != 2 && (flags & IEEE80211_F_SHPREAMBLE))
			txtime += 72 + 24;
		else
			txtime += 144 + 48;
	}

	return txtime;
}

/*
 * rt2870_rx_intr
 */
static void rt2870_rx_intr(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status)
{
	struct rt2870_softc_rx_data *data;
	struct rt2870_softc *sc;
	struct ifnet *ifp;
	int len;
	uint8_t *buf;
	uint32_t dmalen;

	data = priv;
	sc = data->sc;
	ifp = sc->ifp;

	DPRINTFN(sc, 5, ("%s: Rx interrupt: %s\n",
		device_get_nameunit(sc->dev), usbd_errstr(status)));

	if (status != USBD_NORMAL_COMPLETION)
	{
		if (status == USBD_NOT_STARTED || status == USBD_CANCELLED)
			return;

		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall_async(sc->rx_ring.usb_pipe);

		goto skip;
	}

	usbd_get_xfer_status(xfer, NULL, NULL, &len, NULL);

	if (len < RT2870_RX_DESC_SIZE)
	{
		ifp->if_ierrors++;
		goto skip;
	}

	buf = data->buf;

	while (len > (sizeof(uint32_t) + sizeof(struct rt2870_rxinfo)))
	{
		dmalen = (le32toh(*(uint32_t *) buf) & 0xffff);

		if (dmalen == 0 || (dmalen & 3) != 0)
		{
			DPRINTF(sc, ("%s: bad DMA len=%u\n",
				device_get_nameunit(sc->dev), dmalen));
			break;
		}

		if ((dmalen + sizeof(uint32_t) + sizeof(struct rt2870_rxinfo)) > len)
		{
			DPRINTF(sc, ("%s: bad DMA len: DMA len=%u, USB len=%d\n",
			    device_get_nameunit(sc->dev),
				(dmalen + sizeof(uint32_t) + sizeof(struct rt2870_rxinfo)), len));
			break;
		}

		rt2870_rx_frame(sc, buf + sizeof(uint32_t), dmalen);

		buf += (dmalen + sizeof(uint32_t) + sizeof(struct rt2870_rxinfo));
		len -= (dmalen + sizeof(uint32_t) + sizeof(struct rt2870_rxinfo));
	}

skip:

	usbd_setup_xfer(xfer, sc->rx_ring.usb_pipe, data, data->buf,
		RT2870_USB_RX_BULK_BUFLEN, USBD_SHORT_XFER_OK | USBD_NO_COPY,
		USBD_NO_TIMEOUT, rt2870_rx_intr);

	usbd_transfer(xfer);
}

/*
 * rt2870_rx_frame
 */
static void rt2870_rx_frame(struct rt2870_softc *sc,
	uint8_t *buf, uint32_t dmalen)
{
	struct ieee80211com *ic;
	struct ifnet *ifp;
	struct ieee80211_frame *wh;
	struct ieee80211_node *ni;
	struct rt2870_softc_rx_radiotap_header *tap;
	struct rt2870_rxwi *rxwi;
	struct rt2870_rxinfo *rxinfo;
	struct mbuf *m;
	uint32_t rxinfo_flags;
	u_int hdrlen;
	uint8_t cipher_err, rssi, ant, rate;
	int len;

	ic = &sc->ic;
	ifp = ic->ic_ifp;

	/* get Rx wireless info */

	rxwi = (struct rt2870_rxwi *) buf;
	len = (le16toh(rxwi->tid_size) >> RT2870_RXWI_SIZE_SHIFT) &
		RT2870_RXWI_SIZE_MASK;

	if (len > dmalen)
	{
		DPRINTF(sc, ("%s: bad rxwi len: DMA len=%d, rxwi len=%d\n",
			device_get_nameunit(sc->dev), dmalen, len));
		return;
	}

	/* get Rx info */

	rxinfo = (struct rt2870_rxinfo *) ((caddr_t) rxwi + dmalen);
	rxinfo_flags = le32toh(rxinfo->flags);

	DPRINTFN(sc, 5, ("%s: Rx frame: DMA len=%d, len=%d, rxinfo flags=0x%08x\n",
		device_get_nameunit(sc->dev), dmalen, len, rxinfo_flags));
	
	/* check for crc errors */

	if (rxinfo_flags & RT2870_RXINFO_FLAGS_CRC_ERR)
	{
		DPRINTF(sc, ("%s: rxinfo: crc error\n",
			device_get_nameunit(sc->dev)));

		ifp->if_ierrors++;

		if (!(ifp->if_flags & IFF_PROMISC))
			return;
	}

	wh = (struct ieee80211_frame *) (rxwi + 1);

	m = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
	if (m == NULL)
	{
		ifp->if_ierrors++;
		return;
	}

	m_copyback(m, 0, len, (caddr_t) wh);

	m->m_pkthdr.rcvif = ifp;
	m->m_pkthdr.len = m->m_len = len;

	wh = mtod(m, struct ieee80211_frame *);

	/* check for cipher errors */

	if (rxinfo_flags & RT2870_RXINFO_FLAGS_DECRYPTED)
	{
		cipher_err = ((rxinfo_flags >> RT2870_RXINFO_FLAGS_CIPHER_ERR_SHIFT) &
			RT2870_RXINFO_FLAGS_CIPHER_ERR_MASK);
		if (cipher_err == RT2870_RXINFO_FLAGS_CIPHER_ERR_NONE)
		{
			if (wh->i_fc[1] & IEEE80211_FC1_WEP)
				wh->i_fc[1] &= ~IEEE80211_FC1_WEP;

			m->m_flags |= M_WEP;
		}
		else
		{
			DPRINTF(sc, ("%s: rxinfo: cipher error=0x%02x\n",
				device_get_nameunit(sc->dev), cipher_err));

			ifp->if_ierrors++;

			if (!(ifp->if_flags & IFF_PROMISC))
			{
				m_free(m);
				return;
			}
		}
	}
	else
	{
		if (wh->i_fc[1] & IEEE80211_FC1_WEP)
		{
			ifp->if_ierrors++;

			if (!(ifp->if_flags & IFF_PROMISC))
			{
				m_free(m);
				return;
			}
		}
	}

	/* strip L2 padding between IEEE 802.11 frame header and body */

	if (rxinfo_flags & RT2870_RXINFO_FLAGS_L2PAD)
	{
		hdrlen = ieee80211_hdrsize(wh);
		ovbcopy(wh, (caddr_t) wh + 2, hdrlen);
		m_adj(m, 2);
		wh = mtod(m, struct ieee80211_frame *);
	}

	ant = rt2870_maxrssi_rxpath(sc, rxwi);
	rssi = rxwi->rssi[ant];
	rate = rt2870_rxrate(rxwi);

	if (bpf_peers_present(sc->drvbpf))
	{
		tap = &sc->rxtap;

		tap->flags = 0;
		tap->rate = rate;
		tap->chan_freq = htole16(ic->ic_curchan->ic_freq);
		tap->chan_flags = htole16(ic->ic_curchan->ic_flags);
		tap->dbm_antsignal = rt2870_rssi2dbm(sc, rssi, ant);
		tap->antenna = ant;
		tap->antsignal = rssi;

		if (rxwi->bw_mcs & RT2870_RXWI_MCS_SHOTPRE)
			tap->flags |= IEEE80211_RADIOTAP_F_SHORTPRE;

		bpf_mtap2(sc->drvbpf, tap, sc->rxtap_len, m);
	}

	DPRINTFN(sc, 5, ("%s: received frame: ant=%d, rssi=%d/%d/%d, snr=%d/%d, "
		"rate=%d, wcid=0x%02x, key index=%d\n",
		device_get_nameunit(sc->dev),
		ant, rxwi->rssi[0], rxwi->rssi[1], rxwi->rssi[2],
		rxwi->snr[0], rxwi->snr[1],
		rate, rxwi->wcid,
		((rxwi->udf_bssidx_keyidx >> RT2870_RXWI_KEYIDX_SHIFT) & RT2870_RXWI_KEYIDX_MASK)));

	ni = ieee80211_find_rxnode(ic, (struct ieee80211_frame_min *) wh);

	ieee80211_input(ic, m, ni, rssi, RT2870_NOISE_FLOOR, 0);

	ieee80211_free_node(ni);
}

/*
 * rt2870_tx_frame
 */
static int rt2870_tx_frame(struct rt2870_softc *sc,
	struct mbuf *m, struct ieee80211_node *ni, int qid)
{
	struct ieee80211com *ic;
	struct rt2870_softc_tx_ring *ring;
	struct rt2870_softc_tx_data *data;
	struct rt2870_txinfo *txinfo;
	struct rt2870_txwi *txwi;
	struct ieee80211_frame *wh;
	struct rt2870_softc_tx_radiotap_header *tap;
	u_int hdrlen;
	uint8_t type, rate, mcs, pid, wcid;
	uint32_t len;
	uint16_t dur;
	usbd_status error;

	ic = &sc->ic;

	ring = &sc->tx_ring[qid];
	data = &ring->data[ring->cur];
	txinfo = (struct rt2870_txinfo *) data->buf;
	txwi = (struct rt2870_txwi *) (txinfo + 1);

	data->m = m;
	data->ni = ni;

	wh = mtod(m, struct ieee80211_frame *);

	hdrlen = ieee80211_hdrsize(wh);
	type = wh->i_fc[0] & IEEE80211_FC0_TYPE_MASK;

	if (IEEE80211_IS_MULTICAST(wh->i_addr1) || type != IEEE80211_FC0_TYPE_DATA)
		rate = IEEE80211_IS_CHAN_5GHZ(ic->ic_curchan) ? 12 : 2;
	else if (ic->ic_fixed_rate != IEEE80211_FIXED_RATE_NONE)
		rate = ic->ic_fixed_rate;
	else
		rate = ni->ni_rates.rs_rates[ni->ni_txrate];

	rate &= IEEE80211_RATE_VAL;

	mcs = rt2870_rate2mcs(rate);

	len = sizeof(struct rt2870_txwi) + m->m_pkthdr.len;
	len = (len + 3) & ~3; /* align end on a 4-bytes boundary */

	memset(txinfo, 0, sizeof(struct rt2870_txinfo));

	txinfo->len = htole16(len);

	txinfo->qsel_flags |= (RT2870_TXINFO_QSEL_EDCA << RT2870_TXINFO_QSEL_SHIFT);

	pid = mcs + 1;

	wcid = (type == IEEE80211_FC0_TYPE_DATA) ?
		RT2870_AID2WCID(ni->ni_associd) : 0xff;

	memset(txwi, 0, sizeof(struct rt2870_txwi));

	txwi->wcid = wcid;

	txwi->pid_mpdu_len = ((htole16(pid) & RT2870_TXWI_PID_MASK) <<
		RT2870_TXWI_PID_SHIFT) |
		((htole16(m->m_pkthdr.len) & RT2870_TXWI_MPDU_LEN_MASK) <<
		 RT2870_TXWI_MPDU_LEN_SHIFT);

	if (!RT2870_RATE_IS_OFDM(rate))
	{
		txwi->phymode_ifs_stbc_shortgi =
			(RT2870_TXWI_PHYMODE_CCK << RT2870_TXWI_PHYMODE_SHIFT);

		if (rate != 2 && (ic->ic_flags & IEEE80211_F_SHPREAMBLE))
			mcs |= RT2870_TXWI_MCS_SHOTPRE;
	}
	else
	{
		txwi->phymode_ifs_stbc_shortgi =
			(RT2870_TXWI_PHYMODE_OFDM << RT2870_TXWI_PHYMODE_SHIFT);
	}

	txwi->bw_mcs = (RT2870_TXWI_BW_20 << RT2870_TXWI_BW_SHIFT) |
		((mcs & RT2870_TXWI_MCS_MASK) << RT2870_TXWI_MCS_SHIFT);

	/* check if RTS/CTS or CTS-to-self protection is required */

	if (!IEEE80211_IS_MULTICAST(wh->i_addr1) &&
		(m->m_pkthdr.len + IEEE80211_CRC_LEN > ic->ic_rtsthreshold ||
		 ((ic->ic_flags & IEEE80211_F_USEPROT) && RT2870_RATE_IS_OFDM(rate))))
		txwi->txop = (RT2870_TXWI_TXOP_HT << RT2870_TXWI_TXOP_SHIFT);
	else
		txwi->txop = (RT2870_TXWI_TXOP_BACKOFF << RT2870_TXWI_TXOP_SHIFT);

	if (!IEEE80211_IS_MULTICAST(wh->i_addr1))
	{
		txwi->bawin_size_xflags |=
			(RT2870_TXWI_XFLAGS_ACK << RT2870_TXWI_XFLAGS_SHIFT);

		dur = rt2870_txtime(RT2870_ACK_SIZE,
			rt2870_ackrate(ic, rate), ic->ic_flags) + sc->sifs;

		*(uint16_t *) wh->i_dur = htole16(dur);
	}

	/* ask MAC to insert timestamp into probe responses */

	if ((wh->i_fc[0] & (IEEE80211_FC0_TYPE_MASK | IEEE80211_FC0_SUBTYPE_MASK)) ==
		(IEEE80211_FC0_TYPE_MGT | IEEE80211_FC0_SUBTYPE_PROBE_RESP))
		txwi->mpdu_density_flags |=
			(RT2870_TXWI_FLAGS_TS << RT2870_TXWI_FLAGS_SHIFT);

	if (bpf_peers_present(sc->drvbpf))
	{
		tap = &sc->txtap;

		tap->flags = 0;
		tap->rate = rate;
		tap->chan_freq = htole16(ic->ic_curchan->ic_freq);
		tap->chan_flags = htole16(ic->ic_curchan->ic_flags);

		if (mcs & RT2870_TXWI_MCS_SHOTPRE)
			tap->flags |= IEEE80211_RADIOTAP_F_SHORTPRE;

		if (wh->i_fc[1] & IEEE80211_FC1_WEP)
		{
			wh->i_fc[1] &= ~IEEE80211_FC1_WEP;

			bpf_mtap2(sc->drvbpf, tap, sc->txtap_len, m);

			wh->i_fc[1] |= IEEE80211_FC1_WEP;
		}
		else
		{
			bpf_mtap2(sc->drvbpf, tap, sc->txtap_len, m);
		}
	}

	m_copydata(m, 0, m->m_pkthdr.len, (caddr_t) (txwi + 1));

	len += sizeof(struct rt2870_txinfo) + 4;

	DPRINTFN(sc, 5, ("%s: sending frame: qid=%d, len=%d, rate=%d, mcs=%d, wcid=0x%02x, DMA len=%d\n",
		device_get_nameunit(sc->dev),
		qid, m->m_pkthdr.len, rate, mcs, wcid, len));

	usbd_setup_xfer(data->xfer, ring->usb_pipe, data, data->buf, len,
	    USBD_FORCE_SHORT_XFER | USBD_NO_COPY, RT2870_TX_TIMEOUT, rt2870_tx_intr);
	error = usbd_transfer(data->xfer);
	if (error != USBD_NORMAL_COMPLETION && error != USBD_IN_PROGRESS)
	{
		m_freem(m);
		data->m = NULL;
		data->ni = NULL;
		return error;
	}

	ring->queued++;
	ring->cur = (ring->cur + 1) % RT2870_SOFTC_TX_RING_DATA_COUNT;

	return 0;
}

/*
 * rt2870_tx_raw
 */
static int rt2870_tx_raw(struct rt2870_softc *sc,
	struct mbuf *m, struct ieee80211_node *ni,
	const struct ieee80211_bpf_params *params)
{
	return 0;
}

/*
 * rt2870_tx_intr
 */
static void rt2870_tx_intr(usbd_xfer_handle xfer,
	usbd_private_handle priv, usbd_status status)
{
	struct rt2870_softc_tx_data *data;
	struct rt2870_softc	*sc;
	struct rt2870_softc_tx_ring *ring;
	struct ifnet *ifp;

	data = priv;
	sc = data->sc;
	ring = &sc->tx_ring[data->qid];
	ifp = sc->ic.ic_ifp;

	DPRINTFN(sc, 5, ("%s: Tx interrupt: %s\n",
		device_get_nameunit(sc->dev), usbd_errstr(status)));

	if (status != USBD_NORMAL_COMPLETION)
	{
		if (status == USBD_NOT_STARTED || status == USBD_CANCELLED)
			return;

		printf("%s: could not transmit buffer: qid=%d, status=%s\n",
		    device_get_nameunit(sc->dev), data->qid, usbd_errstr(status));

		if (status == USBD_STALLED)
			usbd_clear_endpoint_stall_async(ring->usb_pipe);

		ifp->if_oerrors++;

		return;
	}

	if (data->m->m_flags & M_TXCB)
		ieee80211_process_callback(data->ni, data->m,
			(status == USBD_NORMAL_COMPLETION) ? 0 : ETIMEDOUT);

	m_freem(data->m);

	ieee80211_free_node(data->ni);

	data->m = NULL;
	data->ni = NULL;

	ring->queued--;
	ifp->if_opackets++;

	sc->tx_timer = 0;

	ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;

	rt2870_start(ifp);
}

/*
 * rt2870_alloc_rx_ring
 */
static int rt2870_alloc_rx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_rx_ring *ring)
{
	struct rt2870_softc_rx_data *data;
	int i, error;

	for (i = 0; i < RT2870_SOFTC_RX_RING_DATA_COUNT; i++)
	{
		data = &ring->data[i];

		data->sc = sc;

		data->xfer = usbd_alloc_xfer(sc->usb_dev);
		if (data->xfer == NULL)
		{
			printf("%s: could not allocate Rx xfer\n",
				device_get_nameunit(sc->dev));
			error = ENOMEM;
			goto fail;
		}

		data->buf = usbd_alloc_buffer(data->xfer, RT2870_USB_RX_BULK_BUFLEN);
		if (data->buf == NULL)
		{
			printf("%s: could not allocate Rx buffer\n",
				device_get_nameunit(sc->dev));
			error = ENOMEM;
			goto fail;
		}
	}

	return 0;

fail:
	
	rt2870_free_rx_ring(sc, ring);

	return error;
}

/*
 * rt2870_free_rx_ring
 */
static void rt2870_free_rx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_rx_ring *ring)
{
	struct rt2870_softc_rx_data *data;
	int i;

	for (i = 0; i < RT2870_SOFTC_RX_RING_DATA_COUNT; i++)
	{
		data = &ring->data[i];

		if (data->xfer != NULL)
		{
			usbd_free_xfer(data->xfer);
			data->xfer = NULL;
		}
	}
}

/*
 * rt2870_alloc_tx_ring
 */
static int rt2870_alloc_tx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_tx_ring *ring, int qid)
{
	struct rt2870_softc_tx_data *data;
	int i, error;

	ring->queued = 0;
	ring->cur = 0;

	for (i = 0; i < RT2870_SOFTC_TX_RING_DATA_COUNT; i++)
	{
		data = &ring->data[i];

		data->sc = sc;
		data->qid = qid;

		data->xfer = usbd_alloc_xfer(sc->usb_dev);
		if (data->xfer == NULL)
		{
			printf("%s: could not allocate Tx xfer\n",
			    device_get_nameunit(sc->dev));
			error = ENOMEM;
			goto fail;
		}

		data->buf = usbd_alloc_buffer(data->xfer,
			RT2870_TX_DESC_SIZE + MCLBYTES);
		if (data->buf == NULL)
		{
			printf("%s: could not allocate Tx buffer\n",
			    device_get_nameunit(sc->dev));
			error = ENOMEM;
			goto fail;
		}

		memset(data->buf, 0, RT2870_TX_DESC_SIZE);
	}

	return 0;

fail:
	
	rt2870_free_tx_ring(sc, ring);

	return error;
}

/*
 * rt2870_free_tx_ring
 */
static void rt2870_free_tx_ring(struct rt2870_softc *sc,
	struct rt2870_softc_tx_ring *ring)
{
	struct rt2870_softc_tx_data *data;
	int i;

	for (i = 0; i < RT2870_SOFTC_TX_RING_DATA_COUNT; i++)
	{
		data = &ring->data[i];

		if (data->xfer != NULL)
		{
			usbd_free_xfer(data->xfer);
			data->xfer = NULL;
		}

		if (data->m != NULL)
		{
			m_free(data->m);
			data->m = NULL;
		}

		if (data->ni != NULL)
		{
			ieee80211_free_node(data->ni);
			data->ni = NULL;
		}
	}
}

static device_method_t rt2870_dev_methods[] =
{
	DEVMETHOD(device_probe,	rt2870_probe),
	DEVMETHOD(device_attach, rt2870_attach),
	DEVMETHOD(device_detach, rt2870_detach),
	{ 0, 0 }
};

static driver_t rt2870_driver =
{
	"rt2870",
	rt2870_dev_methods,
	sizeof(struct rt2870_softc),
};

static devclass_t rt2870_dev_class;

DRIVER_MODULE(rt2870, uhub, rt2870_driver, rt2870_dev_class,
	usbd_driver_load, 0);

MODULE_DEPEND(rt2870, wlan, 1, 1, 1);
MODULE_DEPEND(rt2870, wlan_amrr, 1, 1, 1);
MODULE_DEPEND(rt2870, usb, 1, 1, 1);
