
#include "rt2870_read_eeprom.h"
#include "rt2870_eeprom.h"
#include "rt2870_io.h"
#include "rt2870_debug.h"

/*
 * Static function prototypes
 */

static uint32_t rt2870_read_eeprom_b4inc(uint32_t b32, int8_t delta);

/*
 * rt2870_read_eeprom
 */
void rt2870_read_eeprom(struct rt2870_softc *sc)
{
	struct ieee80211com *ic;
	uint16_t val;
	uint32_t tmp;
	int i;

	ic = &sc->ic;

	/* read EEPROM version */

	sc->eeprom_rev = rt2870_io_eeprom_read(sc, RT2870_EEPROM_VERSION);

	DPRINTF(sc, ("%s: EEPROM rev=0x%04x\n",
		device_get_nameunit(sc->dev), sc->eeprom_rev));

	/* read MAC address */

	rt2870_io_eeprom_read_multi(sc, RT2870_EEPROM_ADDRESS,
		ic->ic_myaddr, IEEE80211_ADDR_LEN);

	DPRINTF(sc, ("%s: EEPROM mac address=%02x:%02x:%02x:%02x:%02x:%02x\n",
		device_get_nameunit(sc->dev),
		ic->ic_myaddr[0], ic->ic_myaddr[1], ic->ic_myaddr[2],
		ic->ic_myaddr[3], ic->ic_myaddr[4], ic->ic_myaddr[5]));

	/* read RF information */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_ANTENNA);
	if (val == 0xffff)
	{
		printf("%s: invalid EEPROM antenna info\n",
			device_get_nameunit(sc->dev));
	}
	else
	{
		sc->rf_rev = (val >> 8) & 0xf;
		sc->ntxpath = (val >> 4) & 0xf;
		sc->nrxpath = (val & 0xf);
	}

	DPRINTF(sc, ("%s: EEPROM RF rev=0x%04x, paths=%dT%dR\n",
		device_get_nameunit(sc->dev), sc->rf_rev, sc->ntxpath, sc->nrxpath));

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_CONFIG);

	sc->tx_agc_cntl = (val >> 1) & 0x1;
	sc->ext_2ghz_lna = (val >> 2) & 0x1;
	sc->ext_5ghz_lna = (val >> 3) & 0x1;

	DPRINTF(sc, ("%s: EEPROM NIC config: Tx AGC control=%d, ext LNA gains=%d/%d\n",
		device_get_nameunit(sc->dev),
		sc->tx_agc_cntl, sc->ext_2ghz_lna, sc->ext_5ghz_lna));

	/* read country code */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_COUNTRY);

	sc->country_2ghz = (val >> 8) & 0xff;
	sc->country_5ghz = (val & 0xff);

	DPRINTF(sc, ("%s: EEPROM country code=%d/%d\n",
		device_get_nameunit(sc->dev), sc->country_2ghz, sc->country_5ghz));

	/* read RF frequency offset */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_RF_FREQ_OFF);

	if ((val & 0xff) != 0xff)
	{
		sc->rf_freq_off = (val & 0xff);
	}
	else
	{
		printf("%s: invalid EEPROM RF freq offset\n",
			device_get_nameunit(sc->dev));

		sc->rf_freq_off = 0;
	}

	DPRINTF(sc, ("%s: EEPROM freq offset=0x%02x\n",
		device_get_nameunit(sc->dev), sc->rf_freq_off));

	/* read LEDs operating mode */

	if (((val >> 8) & 0xff) != 0xff)
	{
		sc->led_cntl = ((val >> 8) & 0xff);
		sc->led_off[0] = rt2870_io_eeprom_read(sc, RT2870_EEPROM_LED1_OFF);
		sc->led_off[1] = rt2870_io_eeprom_read(sc, RT2870_EEPROM_LED2_OFF);
		sc->led_off[2] = rt2870_io_eeprom_read(sc, RT2870_EEPROM_LED3_OFF);
	}
	else
	{
		printf("%s: invalid EEPROM LED settings\n",
			device_get_nameunit(sc->dev));

		sc->led_cntl = RT2870_EEPROM_LED_CNTL_DEFAULT;
		sc->led_off[0] = RT2870_EEPROM_LED1_OFF_DEFAULT;
		sc->led_off[1] = RT2870_EEPROM_LED2_OFF_DEFAULT;
		sc->led_off[2] = RT2870_EEPROM_LED3_OFF_DEFAULT;
	}

	DPRINTF(sc, ("%s: EEPROM led cntl=0x%02x, LEDs=0x%04x/0x%04x/0x%04x\n",
		device_get_nameunit(sc->dev), sc->led_cntl,
		sc->led_off[0], sc->led_off[1], sc->led_off[2]));

	/* read RSSI offsets and LNA gains */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_LNA_GAIN);

	sc->lna_gain[0] = (val & 0xff);
	sc->lna_gain[1] = (val >> 8) & 0xff;

	DPRINTF(sc, ("%s: EEPROM LNA gains=0x%02x/0x%02x\n",
		device_get_nameunit(sc->dev), sc->lna_gain[0], sc->lna_gain[1]));

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_RSSI1_OFF_2GHZ);

	sc->rssi_off_2ghz[0] = (val & 0xff);
	sc->rssi_off_2ghz[1] = (val >> 8) & 0xff;

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_RSSI2_OFF_2GHZ);

	sc->rssi_off_2ghz[3] = (val & 0xff);
	sc->lna_gain[2] = (val >> 8) & 0xff;

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_RSSI1_OFF_5GHZ);

	sc->rssi_off_5ghz[0] = (val & 0xff);
	sc->rssi_off_5ghz[1] = (val >> 8) & 0xff;

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_RSSI2_OFF_5GHZ);

	sc->rssi_off_5ghz[3] = (val & 0xff);
	sc->lna_gain[3] = (val >> 8) & 0xff;

	for (i = 2; i < RT2870_SOFTC_LNA_GAIN_COUNT; i++)
	{
		if (sc->lna_gain[i] == 0x00 || sc->lna_gain[i] == 0xff)
		{
			printf("%s: invalid EEPROM LNA gain: %d\n",
				device_get_nameunit(sc->dev), i);

			sc->lna_gain[i] = sc->lna_gain[1];
		}
	}

	for (i = 0; i < RT2870_SOFTC_RSSI_OFF_COUNT; i++)
	{
		if (sc->rssi_off_2ghz[i] < RT2870_EEPROM_RSSI_OFF_MIN ||
			sc->rssi_off_2ghz[i] > RT2870_EEPROM_RSSI_OFF_MAX)
		{
			printf("%s: invalid EEPROM RSSI offset: %d (2GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->rssi_off_2ghz[i] = 0;
		}

		if (sc->rssi_off_5ghz[i] < RT2870_EEPROM_RSSI_OFF_MIN ||
			sc->rssi_off_5ghz[i] > RT2870_EEPROM_RSSI_OFF_MAX)
		{
			printf("%s: invalid EEPROM RSSI offset: %d (5GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->rssi_off_5ghz[i] = 0;
		}
	}

	/* read Tx power settings for 2GHz channels */

	for (i = 0; i < 14; i += 2)
	{
		val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TXPOW_2GHZ_BASE1 + i / 2);

		sc->txpow1[i + 0] = (int8_t) (val & 0xff);
		sc->txpow1[i + 1] = (int8_t) (val >> 8);

		val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TXPOW_2GHZ_BASE2 + i / 2);

		sc->txpow2[i + 0] = (int8_t) (val & 0xff);
		sc->txpow2[i + 1] = (int8_t) (val >> 8);
	}

	/* read Tx power settings for 5GHz channels */

	for (; i < RT2870_SOFTC_TXPOW_COUNT; i += 2)
	{
		val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TXPOW_5GHZ_BASE1 + i / 2);

		sc->txpow1[i + 0] = (int8_t) (val & 0xff);
		sc->txpow1[i + 1] = (int8_t) (val >> 8);

		val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TXPOW_5GHZ_BASE2 + i / 2);

		sc->txpow2[i + 0] = (int8_t) (val & 0xff);
		sc->txpow2[i + 1] = (int8_t) (val >> 8);
	}

	/* fix broken Tx power settings */

    for (i = 0; i < 14; i++)
	{
		if (sc->txpow1[i] < RT2870_EEPROM_TXPOW_2GHZ_MIN ||
			sc->txpow1[i] > RT2870_EEPROM_TXPOW_2GHZ_MAX)
		{
			printf("%s: invalid EEPROM Tx power1: %d (2GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->txpow1[i] = RT2870_EEPROM_TXPOW_2GHZ_DEFAULT;
		}

		if (sc->txpow2[i] < RT2870_EEPROM_TXPOW_2GHZ_MIN ||
			sc->txpow2[i] > RT2870_EEPROM_TXPOW_2GHZ_MAX)
		{
			printf("%s: invalid EEPROM Tx power2: %d (2GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->txpow2[i] = RT2870_EEPROM_TXPOW_2GHZ_DEFAULT;
		}
	}

	for (; i < RT2870_SOFTC_TXPOW_COUNT; i++)
	{
		if (sc->txpow1[i] < RT2870_EEPROM_TXPOW_5GHZ_MIN ||
			sc->txpow1[i] > RT2870_EEPROM_TXPOW_5GHZ_MAX)
		{
			printf("%s: invalid EEPROM Tx power1: %d (5GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->txpow1[i] = RT2870_EEPROM_TXPOW_5GHZ_DEFAULT;
		}

		if (sc->txpow2[i] < RT2870_EEPROM_TXPOW_5GHZ_MIN ||
			sc->txpow2[i] > RT2870_EEPROM_TXPOW_5GHZ_MAX)
		{
			printf("%s: invalid EEPROM Tx power2: %d (5GHz)\n",
				device_get_nameunit(sc->dev), i);

			sc->txpow2[i] = RT2870_EEPROM_TXPOW_5GHZ_DEFAULT;
		}
	}

	/* read Tx power compensation */

	if (sc->tx_agc_cntl)
		sc->tx_agc_cntl_2ghz = sc->tx_agc_cntl_5ghz = 1;

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TXPOW_COMP_DELTA);

	sc->txpow_comp_delta_2ghz = 0;
	sc->txpow_comp_delta_5ghz = 0;

	if ((val & 0xff) != 0xff && (val & 0x80))
	{
		sc->txpow_comp_delta_2ghz = (val & 0xf);
		if (!(val & 0x40))
			sc->txpow_comp_delta_2ghz = -sc->txpow_comp_delta_2ghz;
	}

	val >>= 8;

	if ((val & 0xff) != 0xff && (val & 0x80))
	{
		sc->txpow_comp_delta_5ghz = (val & 0xf);
		if (!(val & 0x40))
			sc->txpow_comp_delta_5ghz = -sc->txpow_comp_delta_5ghz;
	}

	DPRINTF(sc, ("%s: EEPROM Tx power compensation deltas=%d(2MHz), %d(5MHz)\n",
		device_get_nameunit(sc->dev),
		sc->txpow_comp_delta_2ghz, sc->txpow_comp_delta_5ghz));

	for (i = 0; i < RT2870_SOFTC_TXPOW_COMP_COUNT; i++)
	{
		rt2870_io_eeprom_read_multi(sc, RT2870_EEPROM_TXPOW_COMP_BASE + i * sizeof(uint32_t),
			&tmp, sizeof(uint32_t));

		sc->txpow_comp_20mhz[i] = tmp;
		sc->txpow_comp_40mhz_2ghz[i] = rt2870_read_eeprom_b4inc(tmp,
			sc->txpow_comp_delta_2ghz);
		sc->txpow_comp_40mhz_5ghz[i] = rt2870_read_eeprom_b4inc(tmp,
			sc->txpow_comp_delta_5ghz);

		DPRINTF(sc, ("%s: EEPROM Tx power compensation #%d=0x%08x(20MHz), 0x%08x(40MHz/2GHz), 0x%08x(40MHz/5GHz)\n",
			device_get_nameunit(sc->dev), i,
			sc->txpow_comp_20mhz[i], sc->txpow_comp_40mhz_2ghz[i], sc->txpow_comp_40mhz_5ghz[i]));
	}

	/* read factory-calibrated samples for temperature compensation */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_2GHZ_BOUND1);

	sc->tssi_2ghz[0] = (val & 0xff);	/* [-4] */
	sc->tssi_2ghz[1] = (val >> 8);		/* [-3] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_2GHZ_BOUND2);

	sc->tssi_2ghz[2] = (val & 0xff);	/* [-2] */
	sc->tssi_2ghz[3] = (val >> 8);		/* [-1] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_2GHZ_BOUND3);

	sc->tssi_2ghz[4] = (val & 0xff);	/* [0] */
	sc->tssi_2ghz[5] = (val >> 8);		/* [+1] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_2GHZ_BOUND4);

	sc->tssi_2ghz[6] = (val & 0xff);	/* [+2] */
	sc->tssi_2ghz[7] = (val >> 8);		/* [+3] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_2GHZ_BOUND5);

	sc->tssi_2ghz[8] = (val & 0xff);	/* [+4] */
	sc->tssi_step_2ghz = (val >> 8);

	if (sc->tssi_2ghz[4] == 0xff)
		sc->tx_agc_cntl_2ghz = 0;

	DPRINTF(sc, ("%s: EEPROM TSSI 2GHz: 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, "
		"0x%02x, 0x%02x, step=%d\n",
		device_get_nameunit(sc->dev),
		sc->tssi_2ghz[0], sc->tssi_2ghz[1], sc->tssi_2ghz[2],
		sc->tssi_2ghz[3], sc->tssi_2ghz[4], sc->tssi_2ghz[5],
		sc->tssi_2ghz[6], sc->tssi_2ghz[7], sc->tssi_2ghz[8],
		sc->tssi_step_2ghz));

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_5GHZ_BOUND1);

	sc->tssi_5ghz[0] = (val & 0xff);	/* [-4] */
	sc->tssi_5ghz[1] = (val >> 8);		/* [-3] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_5GHZ_BOUND2);

	sc->tssi_5ghz[2] = (val & 0xff);	/* [-2] */
	sc->tssi_5ghz[3] = (val >> 8);		/* [-1] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_5GHZ_BOUND3);

	sc->tssi_5ghz[4] = (val & 0xff);	/* [0] */
	sc->tssi_5ghz[5] = (val >> 8);		/* [+1] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_5GHZ_BOUND4);

	sc->tssi_5ghz[6] = (val & 0xff);	/* [+2] */
	sc->tssi_5ghz[7] = (val >> 8);		/* [+3] */

	val = rt2870_io_eeprom_read(sc, RT2870_EEPROM_TSSI_5GHZ_BOUND5);

	sc->tssi_5ghz[8] = (val & 0xff);	/* [+4] */
	sc->tssi_step_5ghz = (val >> 8);

	if (sc->tssi_5ghz[4] == 0xff)
		sc->tx_agc_cntl_5ghz = 0;

	DPRINTF(sc, ("%s: EEPROM TSSI 5GHz: 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x, "
		"0x%02x, 0x%02x, step=%d\n",
		device_get_nameunit(sc->dev),
		sc->tssi_5ghz[0], sc->tssi_5ghz[1], sc->tssi_5ghz[2],
		sc->tssi_5ghz[3], sc->tssi_5ghz[4], sc->tssi_5ghz[5],
		sc->tssi_5ghz[6], sc->tssi_5ghz[7], sc->tssi_5ghz[8],
		sc->tssi_step_5ghz));

	/* read default BBP settings */

	rt2870_io_eeprom_read_multi(sc, RT2870_EEPROM_BBP_BASE,
		sc->bbp_eeprom, RT2870_SOFTC_BBP_EEPROM_COUNT * sizeof(uint16_t));
}

/*
 * rt2870_read_eeprom_b4inc
 */
static uint32_t rt2870_read_eeprom_b4inc(uint32_t b32, int8_t delta)
{
	int8_t i, b4;

	for (i = 0; i < 8; i++)
	{
		b4 = b32 & 0xf;
		b4 += delta;

		if (b4 < 0)
			b4 = 0;
		else if (b4 > 0xf)
			b4 = 0xf;

		b32 = (b32 >> 4) | (b4 << 28);
	}

	return b32;
}
